"""Protocol engine commands sub-state."""

from __future__ import annotations

import enum
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, Union
from typing_extensions import assert_never

from opentrons_shared_data.errors import EnumeratedError, ErrorCodes, PythonException

from opentrons.ordered_set import OrderedSet

from opentrons.hardware_control.types import DoorState
from opentrons.protocol_engine.actions.actions import (
    ResumeFromRecoveryAction,
    RunCommandAction,
    SetErrorRecoveryPolicyAction,
)
from opentrons.protocol_engine.commands.unsafe.unsafe_ungrip_labware import (
    UnsafeUngripLabwareCommandType,
)
from opentrons.protocol_engine.error_recovery_policy import (
    ErrorRecoveryPolicy,
    ErrorRecoveryType,
)
from opentrons.protocol_engine.notes.notes import CommandNote
from opentrons.protocol_engine.state import update_types

from ..actions import (
    Action,
    QueueCommandAction,
    SucceedCommandAction,
    FailCommandAction,
    PlayAction,
    PauseAction,
    StopAction,
    FinishAction,
    HardwareStoppedAction,
    DoorChangeAction,
)

from ..commands import Command, CommandStatus, CommandIntent, CommandCreate
from ..errors import (
    RunStoppedError,
    ErrorOccurrence,
    RobotDoorOpenError,
    SetupCommandNotAllowedError,
    FixitCommandNotAllowedError,
    ResumeFromRecoveryNotAllowedError,
    PauseNotAllowedError,
    UnexpectedProtocolError,
    ProtocolCommandFailedError,
)
from ..types import EngineStatus
from ._abstract_store import HasState, HandlesActions
from .command_history import (
    CommandEntry,
    CommandHistory,
)
from .config import Config


class QueueStatus(enum.Enum):
    """Execution status of the command queue."""

    SETUP = enum.auto()
    """The engine has been created, but the run has not yet started.

    New protocol commands may be enqueued, but will wait to execute.
    New setup commands may be enqueued and will execute immediately.
    New fixup commands may not be enqueued.
    """

    RUNNING = enum.auto()
    """The queue is running through protocol commands.

    New protocol commands may be enqueued and will execute immediately.
    New setup commands may not be enqueued.
    New fixup commands may not be enqueued.
    """

    PAUSED = enum.auto()
    """Execution of protocol commands has been paused.

    New protocol commands may be enqueued, but will wait to execute.
    New setup commands may not be enqueued.
    New fixup commands may not be enqueued.
    """

    AWAITING_RECOVERY = enum.auto()
    """A protocol command has encountered a recoverable error.

    New protocol commands may be enqueued, but will wait to execute.
    New setup commands may not be enqueued.
    New fixup commands may be enqueued and will execute immediately.
    """

    AWAITING_RECOVERY_PAUSED = enum.auto()
    """Execution of fixit commands has been paused.

    New protocol and fixit commands may be enqueued, but will usually wait to execute.
    There are certain exceptions where fixit commands will still run.

    New setup commands may not be enqueued.
    """


class RunResult(enum.Enum):
    """Result of the run."""

    SUCCEEDED = enum.auto()
    FAILED = enum.auto()
    STOPPED = enum.auto()


@dataclass(frozen=True)
class CommandSlice:
    """A subset of all commands in state."""

    commands: List[Command]
    cursor: int
    total_length: int


@dataclass(frozen=True)
class CommandErrorSlice:
    """A subset of all commands errors in state."""

    commands_errors: List[ErrorOccurrence]
    cursor: int
    total_length: int


@dataclass(frozen=True)
class CommandPointer:
    """Brief info about a command and where to find it."""

    command_id: str
    command_key: str
    created_at: datetime
    index: int


@dataclass(frozen=True)
class _RecoveryTargetInfo:
    """Info about the failed command that we're currently recovering from."""

    command_id: str

    state_update_if_false_positive: update_types.StateUpdate
    """See `CommandView.get_state_update_if_continued()`."""


@dataclass
class CommandState:
    """State of all protocol engine command resources."""

    command_history: CommandHistory

    queue_status: QueueStatus
    """Whether the engine is currently pulling new commands off the queue to execute.

    A command may still be executing, and the robot may still be in motion,
    even if PAUSED.
    """

    run_started_at: Optional[datetime]
    """The time the run was started.

    Set when the first `PlayAction` is dispatched.
    """

    run_completed_at: Optional[datetime]
    """The time the run has completed.

    Set when 'HardwareStoppedAction' is dispatched.
    """

    is_door_blocking: bool
    """Whether the door is open when enable_door_safety_switch feature flag is ON."""

    run_result: Optional[RunResult]
    """Whether the run is done and succeeded, failed, or stopped.

    This doesn't include the post-run finish steps (homing and dropping tips).

    Once set, this status is immutable.
    """

    run_error: Optional[ErrorOccurrence]
    """The run's fatal error occurrence, if there was one.

    Individual command errors, which may or may not be fatal,
    are stored on the individual commands themselves.
    """

    failed_command: Optional[CommandEntry]
    """The most recent command failure, if any."""
    # TODO(mm, 2024-03-19): This attribute is currently only used to help robot-server
    # with pagination, but "the failed command" is an increasingly nuanced idea, now
    # that we're doing error recovery. See if we can implement robot-server pagination
    # atop simpler concepts, like "the last command that ran" or "the next command that
    # would run."
    #
    # TODO(mm, 2024-04-03): Can this be replaced by
    # CommandHistory.get_terminal_command() now?

    command_error_recovery_types: Dict[str, ErrorRecoveryType]
    """For each command that failed (indexed by ID), what its recovery type was.

    This only includes commands that actually failed, not the ones that we mark as
    failed but that are effectively "cancelled" because a command before them failed.

    This separate attribute is a stopgap until error recovery concepts are a bit more
    stable. Eventually, we might want this info to be stored directly on each command.
    """

    recovery_target: Optional[_RecoveryTargetInfo]
    """If we're currently recovering from a command failure, info about that command."""

    finish_error: Optional[ErrorOccurrence]
    """The error that happened during the post-run finish steps (homing & dropping tips), if any."""

    latest_protocol_command_hash: Optional[str]
    """The latest PROTOCOL command hash value received in a QueueCommandAction.

    This value can be used to generate future hashes.
    """

    has_entered_error_recovery: bool
    """Whether the run has entered error recovery."""

    stopped_by_estop: bool
    """If this is set to True, the engine was stopped by an estop event."""

    error_recovery_policy: ErrorRecoveryPolicy
    """See `CommandView.get_error_recovery_policy()`."""


class CommandStore(HasState[CommandState], HandlesActions):
    """Command state container for run-level command concerns."""

    _state: CommandState

    def __init__(
        self,
        *,
        config: Config,
        is_door_open: bool,
        error_recovery_policy: ErrorRecoveryPolicy,
    ) -> None:
        """Initialize a CommandStore and its state."""
        self._config = config
        self._state = CommandState(
            command_history=CommandHistory(),
            queue_status=QueueStatus.SETUP,
            is_door_blocking=is_door_open and config.block_on_door_open,
            run_result=None,
            run_error=None,
            finish_error=None,
            failed_command=None,
            command_error_recovery_types={},
            recovery_target=None,
            run_completed_at=None,
            run_started_at=None,
            latest_protocol_command_hash=None,
            stopped_by_estop=False,
            error_recovery_policy=error_recovery_policy,
            has_entered_error_recovery=False,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        match action:
            case QueueCommandAction():
                self._handle_queue_command_action(action)
            case RunCommandAction():
                self._handle_run_command_action(action)
            case SucceedCommandAction():
                self._handle_succeed_command_action(action)
            case FailCommandAction():
                self._handle_fail_command_action(action)
            case PlayAction():
                self._handle_play_action(action)
            case PauseAction():
                self._handle_pause_action(action)
            case ResumeFromRecoveryAction():
                self._handle_resume_from_recovery_action(action)
            case StopAction():
                self._handle_stop_action(action)
            case FinishAction():
                self._handle_finish_action(action)
            case HardwareStoppedAction():
                self._handle_hardware_stopped_action(action)
            case DoorChangeAction():
                self._handle_door_change_action(action)
            case SetErrorRecoveryPolicyAction():
                self._handle_set_error_recovery_policy_action(action)
            case _:
                pass

    def _handle_queue_command_action(self, action: QueueCommandAction) -> None:
        # TODO(mc, 2021-06-22): mypy has trouble with this automatic
        # request > command mapping, figure out how to type precisely
        # (or wait for a future mypy version that can figure it out).
        queued_command = action.request._CommandCls.model_construct(
            id=action.command_id,
            key=(
                action.request.key
                if action.request.key is not None
                else (action.request_hash or action.command_id)
            ),
            createdAt=action.created_at,
            params=action.request.params,  # type: ignore[arg-type]
            intent=action.request.intent,
            status=CommandStatus.QUEUED,
            failedCommandId=action.failed_command_id,
        )

        self._state.command_history.append_queued_command(queued_command)

        if action.request_hash is not None:
            self._state.latest_protocol_command_hash = action.request_hash

    def _handle_run_command_action(self, action: RunCommandAction) -> None:
        prev_entry = self._state.command_history.get(action.command_id)

        running_command = prev_entry.command.model_copy(
            update={
                "status": CommandStatus.RUNNING,
                "startedAt": action.started_at,
            }
        )

        self._state.command_history.set_command_running(running_command)

    def _handle_succeed_command_action(self, action: SucceedCommandAction) -> None:
        succeeded_command = action.command
        self._state.command_history.set_command_succeeded(succeeded_command)

    def _handle_fail_command_action(self, action: FailCommandAction) -> None:
        prev_entry = self.state.command_history.get(action.command_id)

        if isinstance(action.error, EnumeratedError):  # The error was undefined.
            public_error_occurrence = ErrorOccurrence.from_failed(
                id=action.error_id,
                createdAt=action.failed_at,
                error=action.error,
            )
            # An empty state update, to no-op.
            state_update_if_false_positive = update_types.StateUpdate()
        else:  # The error was defined.
            public_error_occurrence = action.error.public
            state_update_if_false_positive = action.error.state_update_if_false_positive

        self._update_to_failed(
            command_id=action.command_id,
            failed_at=action.failed_at,
            error_occurrence=public_error_occurrence,
            error_recovery_type=action.type,
            notes=action.notes,
        )
        self._state.failed_command = self._state.command_history.get(action.command_id)

        if (
            prev_entry.command.intent in (CommandIntent.PROTOCOL, None)
            and action.type == ErrorRecoveryType.WAIT_FOR_RECOVERY
        ):
            self._state.queue_status = QueueStatus.AWAITING_RECOVERY
            self._state.recovery_target = _RecoveryTargetInfo(
                command_id=action.command_id,
                state_update_if_false_positive=state_update_if_false_positive,
            )
            self._state.has_entered_error_recovery = True

        # When one command fails, we generally also cancel the commands that
        # would have been queued after it.
        other_command_ids_to_fail: List[str]
        if prev_entry.command.intent == CommandIntent.SETUP:
            other_command_ids_to_fail = list(
                self._state.command_history.get_setup_queue_ids()
            )
        elif prev_entry.command.intent == CommandIntent.FIXIT:
            other_command_ids_to_fail = list(
                self._state.command_history.get_fixit_queue_ids()
            )
        elif (
            prev_entry.command.intent == CommandIntent.PROTOCOL
            or prev_entry.command.intent is None
        ):
            if action.type == ErrorRecoveryType.FAIL_RUN:
                other_command_ids_to_fail = list(
                    self._state.command_history.get_queue_ids()
                )
            elif (
                action.type == ErrorRecoveryType.WAIT_FOR_RECOVERY
                or action.type == ErrorRecoveryType.CONTINUE_WITH_ERROR
                or action.type == ErrorRecoveryType.ASSUME_FALSE_POSITIVE_AND_CONTINUE
            ):
                other_command_ids_to_fail = []
            else:
                assert_never(action.type)
        else:
            assert_never(prev_entry.command.intent)
        for command_id in other_command_ids_to_fail:
            # TODO(mc, 2022-06-06): add new "cancelled" status or similar
            self._update_to_failed(
                command_id=command_id,
                failed_at=action.failed_at,
                error_occurrence=None,
                error_recovery_type=None,
                notes=None,
            )

    def _handle_play_action(self, action: PlayAction) -> None:
        if not self._state.run_result:
            self._state.run_started_at = (
                self._state.run_started_at or action.requested_at
            )
            match self._state.queue_status:
                case QueueStatus.SETUP:
                    self._state.queue_status = (
                        QueueStatus.PAUSED
                        if self._state.is_door_blocking
                        else QueueStatus.RUNNING
                    )
                case QueueStatus.AWAITING_RECOVERY_PAUSED:
                    self._state.queue_status = QueueStatus.AWAITING_RECOVERY
                case QueueStatus.PAUSED:
                    self._state.queue_status = QueueStatus.RUNNING
                case QueueStatus.RUNNING | QueueStatus.AWAITING_RECOVERY:
                    # Nothing for the play action to do. No-op.
                    pass

    def _handle_pause_action(self, action: PauseAction) -> None:
        self._state.queue_status = QueueStatus.PAUSED

    def _handle_resume_from_recovery_action(
        self, action: ResumeFromRecoveryAction
    ) -> None:
        self._state.queue_status = QueueStatus.RUNNING
        self._state.recovery_target = None

    def _handle_stop_action(self, action: StopAction) -> None:
        if not self._state.run_result:
            self._state.recovery_target = None
            self._state.queue_status = QueueStatus.PAUSED

            if action.from_estop:
                self._state.stopped_by_estop = True
                self._state.run_result = RunResult.FAILED
            else:
                self._state.run_result = RunResult.STOPPED

    def _handle_finish_action(self, action: FinishAction) -> None:
        if not self._state.run_result:
            self._state.recovery_target = None
            self._state.queue_status = QueueStatus.PAUSED

            if action.set_run_status:
                self._state.run_result = (
                    RunResult.SUCCEEDED
                    if not action.error_details
                    else RunResult.FAILED
                )
            else:
                self._state.run_result = RunResult.STOPPED

            if not self._state.run_error and action.error_details:
                self._state.run_error = self._map_run_exception_to_error_occurrence(
                    action.error_details.error_id,
                    action.error_details.created_at,
                    action.error_details.error,
                )
        else:
            # HACK(sf): There needs to be a better way to set
            # an estop error than this else clause
            if self._state.stopped_by_estop and action.error_details:
                self._state.run_error = self._map_run_exception_to_error_occurrence(
                    action.error_details.error_id,
                    action.error_details.created_at,
                    action.error_details.error,
                )

    def _handle_hardware_stopped_action(self, action: HardwareStoppedAction) -> None:
        self._state.queue_status = QueueStatus.PAUSED
        self._state.run_result = self._state.run_result or RunResult.STOPPED
        self._state.run_completed_at = (
            self._state.run_completed_at or action.completed_at
        )

        if action.finish_error_details:
            self._state.finish_error = self._map_finish_exception_to_error_occurrence(
                action.finish_error_details.error_id,
                action.finish_error_details.created_at,
                action.finish_error_details.error,
            )

    def _handle_door_change_action(self, action: DoorChangeAction) -> None:
        if self._config.block_on_door_open:
            if action.door_state == DoorState.OPEN:
                self._state.is_door_blocking = True
                match self._state.queue_status:
                    case QueueStatus.SETUP:
                        pass
                    case QueueStatus.RUNNING | QueueStatus.PAUSED:
                        self._state.queue_status = QueueStatus.PAUSED
                    case (
                        QueueStatus.AWAITING_RECOVERY
                        | QueueStatus.AWAITING_RECOVERY_PAUSED
                    ):
                        self._state.queue_status = QueueStatus.AWAITING_RECOVERY_PAUSED
            elif action.door_state == DoorState.CLOSED:
                self._state.is_door_blocking = False

    def _handle_set_error_recovery_policy_action(
        self, action: SetErrorRecoveryPolicyAction
    ) -> None:
        self._state.error_recovery_policy = action.error_recovery_policy

    def _update_to_failed(
        self,
        command_id: str,
        failed_at: datetime,
        error_occurrence: Optional[ErrorOccurrence],
        error_recovery_type: Optional[ErrorRecoveryType],
        notes: Optional[List[CommandNote]],
    ) -> None:
        prev_entry = self._state.command_history.get(command_id)
        failed_command = prev_entry.command.model_copy(
            update={
                "completedAt": failed_at,
                "status": CommandStatus.FAILED,
                **({"error": error_occurrence} if error_occurrence is not None else {}),
                # Assume we're not overwriting any existing notes because they can
                # only be added when a command completes, and if we're failing this
                # command, it wouldn't have completed before now.
                **({"notes": notes} if notes is not None else {}),
            }
        )
        self._state.command_history.set_command_failed(failed_command)
        if error_recovery_type is not None:
            self._state.command_error_recovery_types[command_id] = error_recovery_type

    @staticmethod
    def _map_run_exception_to_error_occurrence(
        error_id: str, created_at: datetime, exception: Exception
    ) -> ErrorOccurrence:
        """Map a fatal exception from the main part of the run to an ErrorOccurrence."""
        if (
            isinstance(exception, ProtocolCommandFailedError)
            and exception.original_error is not None
        ):
            return exception.original_error
        elif isinstance(exception, EnumeratedError):
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=exception
            )
        else:
            enumerated_wrapper = UnexpectedProtocolError(
                message=str(exception),
                wrapping=[exception],
            )
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=enumerated_wrapper
            )

    @staticmethod
    def _map_finish_exception_to_error_occurrence(
        error_id: str, created_at: datetime, exception: Exception
    ) -> ErrorOccurrence:
        """Map a fatal exception from the finish phase (drop tip & home) to an ErrorOccurrence."""
        if isinstance(exception, EnumeratedError):
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=exception
            )
        else:
            enumerated_wrapper = PythonException(exc=exception)
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=enumerated_wrapper
            )


class CommandView:
    """Read-only command state view."""

    _state: CommandState

    def __init__(self, state: CommandState) -> None:
        """Initialize the view of command state with its underlying data."""
        self._state = state

    def get(self, command_id: str) -> Command:
        """Get a command by its unique identifier."""
        return self._state.command_history.get(command_id).command

    def get_all(self) -> List[Command]:
        """Get a list of all commands in state.

        Entries are returned in the order of first-added command to last-added command.
        Replacing a command (to change its status, for example) keeps its place in the
        ordering.
        """
        return self._state.command_history.get_all_commands()

    def get_slice(
        self, cursor: Optional[int], length: int, include_fixit_commands: bool
    ) -> CommandSlice:
        """Get a subset of commands around a given cursor.

        If the cursor is omitted, a cursor will be selected automatically
        based on the currently running or most recently executed command,
        and the slice of commands returned is the previous `length` commands
        inclusive of the currently running or most recently executed command.
        """
        command_ids = self._state.command_history.get_filtered_command_ids(
            include_fixit_commands=include_fixit_commands
        )
        total_length = len(command_ids)

        if cursor is None:
            current_pointer = self.get_current()

            if current_pointer is not None:
                cursor = current_pointer.index
            else:
                cursor = total_length - 1

            cursor = max(cursor - length + 1, 0)

        # start is inclusive, stop is exclusive
        start = max(0, min(cursor, total_length - 1))
        stop = min(total_length, start + length)
        commands = self._state.command_history.get_slice(
            start=start, stop=stop, command_ids=command_ids
        )

        return CommandSlice(
            commands=commands,
            cursor=start,
            total_length=total_length,
        )

    def get_errors_slice(
        self,
        cursor: int,
        length: int,
    ) -> CommandErrorSlice:
        """Get a subset of commands error around a given cursor."""
        # start is inclusive, stop is exclusive
        all_errors = self.get_all_errors()
        total_length = len(all_errors)
        actual_cursor = max(0, min(cursor, total_length - 1))
        stop = min(total_length, actual_cursor + length)

        sliced_errors = all_errors[actual_cursor:stop]

        return CommandErrorSlice(
            commands_errors=sliced_errors,
            cursor=actual_cursor,
            total_length=total_length,
        )

    def get_error(self) -> Optional[ErrorOccurrence]:
        """Get the run's fatal error, if there was one."""
        run_error = self._state.run_error
        finish_error = self._state.finish_error

        if run_error and finish_error:
            combined_error = ErrorOccurrence(
                id=finish_error.id,
                createdAt=finish_error.createdAt,
                errorType="RunAndFinishFailed",
                detail=(
                    "The run had a fatal error,"
                    " and another error happened while doing post-run cleanup."
                ),
                # TODO(mm, 2023-07-31): Consider adding a low-priority error code so clients can
                # deemphasize this root node, in favor of its children in wrappedErrors.
                errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                wrappedErrors=[
                    run_error,
                    finish_error,
                ],
            )
            return combined_error
        else:
            return run_error or finish_error

    def get_all_errors(self) -> List[ErrorOccurrence]:
        """Get the run's full error list, if there was none, returns an empty list."""
        failed_commands = self._state.command_history.get_all_failed_commands()
        return [
            command_error.error
            for command_error in failed_commands
            if command_error.error is not None
        ]

    def get_has_entered_recovery_mode(self) -> bool:
        """Get whether the run has entered recovery mode."""
        return self._state.has_entered_error_recovery

    def get_running_command_id(self) -> Optional[str]:
        """Return the ID of the command that's currently running, if there is one."""
        running_command = self._state.command_history.get_running_command()
        if running_command is not None:
            return running_command.command.id
        else:
            return None

    def get_queue_ids(self) -> OrderedSet[str]:
        """Get the IDs of all queued protocol commands, in FIFO order."""
        return self._state.command_history.get_queue_ids()

    def get_current(self) -> Optional[CommandPointer]:
        """Return the "current" command, if any.

        The "current" command is the command that is currently executing,
        or the most recent command to have completed.
        """
        running_command = self._state.command_history.get_running_command()
        if running_command:
            return CommandPointer(
                command_id=running_command.command.id,
                command_key=running_command.command.key,
                created_at=running_command.command.createdAt,
                index=running_command.index,
            )

        most_recently_finalized_command = self.get_most_recently_finalized_command()
        if most_recently_finalized_command:
            return CommandPointer(
                command_id=most_recently_finalized_command.command.id,
                command_key=most_recently_finalized_command.command.key,
                created_at=most_recently_finalized_command.command.createdAt,
                index=most_recently_finalized_command.index,
            )

        return None

    def get_next_to_execute(self) -> Optional[str]:
        """Return the next command in line to be executed.

        Returns:
            The ID of the earliest queued command, if any.

        Raises:
            RunStoppedError: The engine is currently stopped or stopping,
                so it will never run any more commands.
        """
        if self._state.run_result:
            raise RunStoppedError("Engine was stopped")

        # if queue is in recovery mode, return the next fixit command
        next_fixit_cmd = self._state.command_history.get_fixit_queue_ids().head(None)
        if next_fixit_cmd and self._state.queue_status == QueueStatus.AWAITING_RECOVERY:
            return next_fixit_cmd
        if (
            next_fixit_cmd
            and self._state.queue_status == QueueStatus.AWAITING_RECOVERY_PAUSED
            and self._may_run_with_door_open(fixit_command=self.get(next_fixit_cmd))
        ):
            return next_fixit_cmd

        # if there is a setup command queued, prioritize it
        next_setup_cmd = self._state.command_history.get_setup_queue_ids().head(None)
        if (
            self._state.queue_status
            not in [QueueStatus.PAUSED, QueueStatus.AWAITING_RECOVERY]
            and next_setup_cmd
        ):
            return next_setup_cmd

        # if the queue is running, return the next protocol command
        if self._state.queue_status == QueueStatus.RUNNING:
            return self._state.command_history.get_queue_ids().head(None)

        # otherwise we've got nothing to do
        return None

    def get_is_okay_to_clear(self) -> bool:
        """Get whether the engine is stopped or sitting idly so it could be removed."""
        if self.get_is_stopped():
            return True
        elif (
            self.get_status() == EngineStatus.IDLE
            and self._state.command_history.get_running_command() is None
            and len(self._state.command_history.get_setup_queue_ids()) == 0
        ):
            return True
        else:
            return False

    def get_is_door_blocking(self) -> bool:
        """Get whether the robot door is open when 'pause on door open' ff is True."""
        return self._state.is_door_blocking

    def get_is_running(self) -> bool:
        """Get whether the protocol is running & queued commands should be executed."""
        return self._state.queue_status == QueueStatus.RUNNING

    def get_most_recently_finalized_command(self) -> Optional[CommandEntry]:
        """Get the most recent command that has reached its final `status`. See get_command_is_final."""
        run_requested_to_stop = self._state.run_result is not None

        if run_requested_to_stop:
            tail_command = self._state.command_history.get_tail_command()
            if not tail_command:
                return None
            if tail_command.command.status != CommandStatus.RUNNING:
                return tail_command
            else:
                return self._state.command_history.get_prev(tail_command.command.id)
        else:
            most_recently_finalized = (
                self._state.command_history.get_most_recently_completed_command()
            )
            # This iteration is effectively O(1) as we'll only ever have to iterate one or two times at most.
            while most_recently_finalized is not None:
                next_command = self._state.command_history.get_next(
                    most_recently_finalized.command.id
                )
                if (
                    next_command is not None
                    and next_command.command.status != CommandStatus.QUEUED
                    and next_command.command.status != CommandStatus.RUNNING
                ):
                    most_recently_finalized = next_command
                else:
                    break

            return most_recently_finalized

    def get_command_is_final(self, command_id: str) -> bool:
        """Get whether a given command has reached its final `status`.

        This happens when one of the following is true:

        - Its status is `CommandStatus.SUCCEEDED`.
        - Its status is `CommandStatus.FAILED`.
        - Its status is `CommandStatus.QUEUED` but the run has been requested to stop,
          so the run will never reach it.

        Arguments:
            command_id: Command to check.
        """
        status = self.get(command_id).status

        run_requested_to_stop = self._state.run_result is not None

        return (
            status == CommandStatus.SUCCEEDED
            or status == CommandStatus.FAILED
            or (status == CommandStatus.QUEUED and run_requested_to_stop)
        )

    def get_all_commands_final(self) -> bool:
        """Get whether all commands added so far have reached their final `status`.

        See `get_command_is_final()`.

        Raises:
            CommandExecutionFailedError: if any added command failed, and its `intent` wasn't
            `setup`.
        """
        no_command_running = self._state.command_history.get_running_command() is None
        run_requested_to_stop = self._state.run_result is not None
        no_command_to_execute = (
            run_requested_to_stop
            # TODO(mm, 2024-03-15): This ignores queued setup commands,
            # which seems questionable?
            or len(self._state.command_history.get_queue_ids()) == 0
        )

        return no_command_running and no_command_to_execute

    def get_recovery_target(self) -> Optional[CommandPointer]:
        """Return the command currently undergoing error recovery, if any."""
        recovery_target = self._state.recovery_target
        if recovery_target is None:
            return None
        else:
            entry = self._state.command_history.get(recovery_target.command_id)
            return CommandPointer(
                command_id=entry.command.id,
                command_key=entry.command.key,
                created_at=entry.command.createdAt,
                index=entry.index,
            )

    def get_recovery_in_progress_for_command(self, command_id: str) -> bool:
        """Return whether the given command failed and its error recovery is in progress."""
        pointer = self.get_recovery_target()
        return pointer is not None and pointer.command_id == command_id

    def raise_fatal_command_error(self) -> None:
        """Raise the run's fatal command error, if there was one, as an exception.

        The "fatal command error" is the error from any non-setup command.
        It's intended to be used as the fatal error of the overall run
        (see `ProtocolEngine.finish()`) for JSON and live HTTP protocols.

        This isn't useful for Python protocols, which have to account for the
        fatal error of the overall run coming from anywhere in the Python script,
        including in between commands.
        """
        failed_command = self._state.failed_command
        if (
            failed_command
            and failed_command.command.error
            and failed_command.command.intent != CommandIntent.SETUP
        ):
            raise ProtocolCommandFailedError(
                original_error=failed_command.command.error,
                message=failed_command.command.error.detail,
            )

    def get_error_recovery_type(self, command_id: str) -> ErrorRecoveryType:
        """Return the error recovery type with which the given command failed.

        The command ID is assumed to point to a failed command.
        """
        return self._state.command_error_recovery_types[command_id]

    def get_is_stopped(self) -> bool:
        """Get whether an engine stop has completed."""
        return self._state.run_completed_at is not None

    def get_is_stopped_by_estop(self) -> bool:
        """Return whether the engine was stopped specifically by an E-stop."""
        return self._state.stopped_by_estop

    def has_been_played(self) -> bool:
        """Get whether engine has started."""
        return self._state.run_started_at is not None

    def get_is_terminal(self) -> bool:
        """Get whether engine is in a terminal state."""
        return self._state.run_result is not None

    def validate_action_allowed(  # noqa: C901
        self,
        action: Union[
            PlayAction,
            PauseAction,
            StopAction,
            ResumeFromRecoveryAction,
            QueueCommandAction,
        ],
    ) -> Union[
        PlayAction,
        PauseAction,
        StopAction,
        ResumeFromRecoveryAction,
        QueueCommandAction,
    ]:
        """Validate whether a given control action is allowed.

        Returns:
            The action, if valid.

        Raises:
            RunStoppedError: The engine has been stopped.
            RobotDoorOpenError: Cannot resume because the front door is open.
            PauseNotAllowedError: The engine is not running, so cannot be paused.
            SetupCommandNotAllowedError: The engine is running, so a setup command
                may not be added.
        """
        if self._state.run_result is not None:
            raise RunStoppedError("The run has already stopped.")

        elif isinstance(action, PlayAction):
            if self.get_status() in (
                EngineStatus.BLOCKED_BY_OPEN_DOOR,
                EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR,
            ):
                raise RobotDoorOpenError("Front door or top window is currently open.")
            else:
                return action

        elif isinstance(action, PauseAction):
            if not self.get_is_running():
                raise PauseNotAllowedError("Cannot pause a run that is not running.")
            elif self.get_status() == EngineStatus.AWAITING_RECOVERY:
                raise PauseNotAllowedError("Cannot pause a run in recovery mode.")
            else:
                return action

        elif isinstance(action, QueueCommandAction):
            if (
                action.request.intent == CommandIntent.SETUP
                and self._state.queue_status != QueueStatus.SETUP
            ):
                raise SetupCommandNotAllowedError(
                    "Setup commands are not allowed after run has started."
                )
            elif action.request.intent == CommandIntent.FIXIT:
                if self.get_status() == EngineStatus.AWAITING_RECOVERY:
                    return action
                elif self.get_status() in (
                    EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR,
                    EngineStatus.AWAITING_RECOVERY_PAUSED,
                ):
                    if self._may_run_with_door_open(fixit_command=action.request):
                        return action
                    else:
                        raise FixitCommandNotAllowedError(
                            f"{action.request.commandType} fixit command may not run"
                            " until the door is closed and the run is played again."
                        )
                else:
                    raise FixitCommandNotAllowedError(
                        "Fixit commands are not allowed when the run is not in a recoverable state."
                    )
            else:
                return action

        elif isinstance(action, ResumeFromRecoveryAction):
            if self.get_status() != EngineStatus.AWAITING_RECOVERY:
                raise ResumeFromRecoveryNotAllowedError(
                    "Cannot resume from recovery if the run is not in recovery mode."
                )
            elif (
                self.get_status() == EngineStatus.AWAITING_RECOVERY
                and len(self._state.command_history.get_fixit_queue_ids()) > 0
            ):
                raise ResumeFromRecoveryNotAllowedError(
                    "Cannot resume from recovery while there are fixit commands in the queue."
                )
            else:
                return action

        elif isinstance(action, StopAction):
            return action

        else:
            assert_never(action)

    def get_status(self) -> EngineStatus:  # noqa: C901
        """Get the current execution status of the engine."""
        if self._state.run_result:
            # The main part of the run is over, or will be over soon.
            # Have we also completed the post-run finish steps (homing and dropping tips)?
            if self.get_is_stopped():
                # Post-run finish steps have completed. Calculate the engine's final status,
                # taking into account any failures in the run or the post-run finish steps.
                if (
                    self._state.run_result == RunResult.FAILED
                    or self._state.finish_error is not None
                ):
                    return EngineStatus.FAILED
                elif self._state.run_result == RunResult.SUCCEEDED:
                    return EngineStatus.SUCCEEDED
                else:
                    return EngineStatus.STOPPED
            else:
                # Post-run finish steps have not yet completed,
                # and we may even still be executing commands.
                return (
                    EngineStatus.STOP_REQUESTED
                    if self._state.run_result == RunResult.STOPPED
                    else EngineStatus.FINISHING
                )

        elif self._state.queue_status == QueueStatus.RUNNING:
            return EngineStatus.RUNNING

        elif self._state.queue_status == QueueStatus.PAUSED:
            if self._state.is_door_blocking:
                return EngineStatus.BLOCKED_BY_OPEN_DOOR
            else:
                return EngineStatus.PAUSED

        elif self._state.queue_status == QueueStatus.AWAITING_RECOVERY_PAUSED:
            if self._state.is_door_blocking:
                return EngineStatus.AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR
            else:
                return EngineStatus.AWAITING_RECOVERY_PAUSED

        elif self._state.queue_status == QueueStatus.AWAITING_RECOVERY:
            return EngineStatus.AWAITING_RECOVERY

        # todo(mm, 2024-03-19): Does this intentionally return idle if QueueStatus is
        # SETUP and we're currently a setup command?
        return EngineStatus.IDLE

    def get_latest_protocol_command_hash(self) -> Optional[str]:
        """Get the command hash of the last queued command, if any."""
        return self._state.latest_protocol_command_hash

    def get_error_recovery_policy(self) -> ErrorRecoveryPolicy:
        """Return the run's current error recovery policy (see `ErrorRecoveryPolicy`).

        This error recovery policy is not ever evaluated by
        `CommandStore`/`CommandView`. It's stored here for convenience, but evaluated by
        higher-level code.
        """
        return self._state.error_recovery_policy

    def get_state_update_for_false_positive(self) -> update_types.StateUpdate:
        """Return the state update for if the current recovery target was a false positive.

        If we're currently in error recovery mode, and you have decided that the
        underlying command error was a false positive, this returns a state update
        that will undo the error's effects on engine state.
        See `ProtocolEngine.resume_from_recovery(reconcile_false_positive=True)`.
        """
        if self._state.recovery_target is None:
            return update_types.StateUpdate()  # Empty/no-op.
        else:
            return self._state.recovery_target.state_update_if_false_positive

    def _may_run_with_door_open(
        self, *, fixit_command: Command | CommandCreate
    ) -> bool:
        """Return whether the given fixit command is exempt from the usual open-door auto pause.

        This is required for certain error recovery flows, where we want the robot to
        do stuff while the door is open.
        """
        # CommandIntent.PROTOCOL and CommandIntent.SETUP have their own rules for whether
        # they run while the door is open. Passing one of those commands to this function
        # is probably a mistake in the caller's logic.
        assert fixit_command.intent == CommandIntent.FIXIT

        # This type annotation is to make sure the string constant stays in sync and isn't typo'd.
        required_command_type: UnsafeUngripLabwareCommandType = "unsafe/ungripLabware"
        # todo(mm, 2024-10-04): Instead of allowlisting command types, maybe we should
        # add a `mayRunWithDoorOpen: bool` field to command requests.
        return fixit_command.commandType == required_command_type



"""Commands that are "unsafe".

"Unsafe" means that they can cause inaccuracy or incorrect behavior. They should
therefore never be used in protocols, and should only be used otherwise as a last
resort.

These exist as a necessary evil for implementing things like error recovery.
Even in those narrow contexts, these commands must be used with care.
e.g. after an `UpdatePositionEstimators` command, there must be a `Home` command,
or positioning will be subtly wrong. Each unsafe command should document its intended
use case and its caveats.

Because we don't expect unsafe commands to be used in any protocols whose behavior we
must preserve, we may change the commands' semantics over time. We may also change
their shapes if we're confident that it won't break something in robot-server's
persistent storage.
"""

from .unsafe_blow_out_in_place import (
    UnsafeBlowOutInPlaceCommandType,
    UnsafeBlowOutInPlaceParams,
    UnsafeBlowOutInPlaceResult,
    UnsafeBlowOutInPlace,
    UnsafeBlowOutInPlaceCreate,
)
from .unsafe_drop_tip_in_place import (
    UnsafeDropTipInPlaceCommandType,
    UnsafeDropTipInPlaceParams,
    UnsafeDropTipInPlaceResult,
    UnsafeDropTipInPlace,
    UnsafeDropTipInPlaceCreate,
)

from .update_position_estimators import (
    UpdatePositionEstimatorsCommandType,
    UpdatePositionEstimatorsParams,
    UpdatePositionEstimatorsResult,
    UpdatePositionEstimators,
    UpdatePositionEstimatorsCreate,
)

from .unsafe_engage_axes import (
    UnsafeEngageAxesCommandType,
    UnsafeEngageAxesParams,
    UnsafeEngageAxesResult,
    UnsafeEngageAxes,
    UnsafeEngageAxesCreate,
)

from .unsafe_ungrip_labware import (
    UnsafeUngripLabwareCommandType,
    UnsafeUngripLabwareParams,
    UnsafeUngripLabwareResult,
    UnsafeUngripLabware,
    UnsafeUngripLabwareCreate,
)


from .unsafe_place_labware import (
    UnsafePlaceLabwareCommandType,
    UnsafePlaceLabwareParams,
    UnsafePlaceLabwareResult,
    UnsafePlaceLabware,
    UnsafePlaceLabwareCreate,
)


__all__ = [
    # Unsafe blow-out-in-place command models
    "UnsafeBlowOutInPlaceCommandType",
    "UnsafeBlowOutInPlaceParams",
    "UnsafeBlowOutInPlaceResult",
    "UnsafeBlowOutInPlace",
    "UnsafeBlowOutInPlaceCreate",
    # Unsafe drop-tip command models
    "UnsafeDropTipInPlaceCommandType",
    "UnsafeDropTipInPlaceParams",
    "UnsafeDropTipInPlaceResult",
    "UnsafeDropTipInPlace",
    "UnsafeDropTipInPlaceCreate",
    # Update position estimate command models
    "UpdatePositionEstimatorsCommandType",
    "UpdatePositionEstimatorsParams",
    "UpdatePositionEstimatorsResult",
    "UpdatePositionEstimators",
    "UpdatePositionEstimatorsCreate",
    # Unsafe engage axes
    "UnsafeEngageAxesCommandType",
    "UnsafeEngageAxesParams",
    "UnsafeEngageAxesResult",
    "UnsafeEngageAxes",
    "UnsafeEngageAxesCreate",
    # Unsafe ungrip labware
    "UnsafeUngripLabwareCommandType",
    "UnsafeUngripLabwareParams",
    "UnsafeUngripLabwareResult",
    "UnsafeUngripLabware",
    "UnsafeUngripLabwareCreate",
    # Unsafe place labware
    "UnsafePlaceLabwareCommandType",
    "UnsafePlaceLabwareParams",
    "UnsafePlaceLabwareResult",
    "UnsafePlaceLabware",
    "UnsafePlaceLabwareCreate",
]



NUMBER_OF_ZONES = 10
NUMBER_OF_BINS = 128


def validate_histogram_frame(data: bytes, next_frame_id: int) -> bool:
    """Validate Histogram frame, Raise error if invalid."""
    start_delimn = data[0]
    assert (
        start_delimn == 0x81  # histogram start byte
    ), f"Invalid delimn, {hex(start_delimn)} != 0x81."
    frame_id = data[4]
    assert (
        next_frame_id == frame_id
    ), f"Invalid frame id, expected {next_frame_id} got {frame_id}."
    frame_len = data[5]
    assert (
        frame_len == 128  # len is always 128
    ), f"Invalid frame length, expected 128 got {frame_len}."
    return True



from typing_extensions import Protocol
from .types import CalibrationType

from ..util import DeckTransformState


class Calibratable(Protocol[CalibrationType]):
    """Protocol specifying calibration information"""

    @property
    def robot_calibration(self) -> CalibrationType:
        """The currently-active robot calibration of the machine."""
        ...

    def reset_robot_calibration(self) -> None:
        """Reset the active robot calibration to the machine default.

        This may be an identity on some machines but not on others; this
        method is therefore preferred to using set_robot_calibration() with a
        caller-constructed identity.
        """
        ...

    def reset_deck_calibration(self) -> None:
        """Resets only deck calibration data."""
        ...

    def load_deck_calibration(self) -> None:
        """Loads only any deck calibration data that is stored."""
        ...

    def set_robot_calibration(self, robot_calibration: CalibrationType) -> None:
        """Set the current robot calibration from stored data."""
        ...

    def validate_calibration(self) -> DeckTransformState:
        """Check whether the current calibration is valid."""
        ...

    def build_temporary_identity_calibration(self) -> CalibrationType:
        """
        Get temporary default calibration data suitable for use during
        calibration
        """
        ...



"""Helpers for flagging unsafe movements to a Thermocycler Module."""
from typing import Optional

from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules import Thermocycler as HardwareThermocycler


from opentrons.protocol_engine.state.module_substates import ThermocyclerModuleId
from ..types import ModuleLocation, LabwareLocation
from ..state.state import StateStore
from ..errors import ThermocyclerNotOpenError, WrongModuleTypeError

from .equipment import EquipmentHandler


class ThermocyclerMovementFlagger:
    """A helper for flagging unsafe movements to a Thermocycler Module.

    This is only intended for use by movement handlers.
    It's a separate class for independent testability.
    """

    def __init__(
        self,
        state_store: StateStore,
        hardware_api: HardwareControlAPI,
        equipment: EquipmentHandler,
    ) -> None:
        """Initialize the ThermocyclerMovementFlagger.

        Args:
            state_store: The Protocol Engine state store interface. Used to figure out
                         which Thermocycler a labware is in, if any.
            hardware_api: The underlying hardware interface. Used to query
                          Thermocyclers' current lid states.
            equipment: The protocol engine interface to move a present thermocycler to an
                            operable state if need be.
        """
        self._state_store = state_store
        self._hardware_api = hardware_api
        self._equipment = equipment

    async def _verify_tc_lid_status(self, module_id: str) -> None:
        """Ensure the thermocycler's lid state is correct or raise an error."""
        try:
            hw_tc_lid_status = await self._get_hardware_thermocycler_lid_status(
                module_id=module_id
            )
        except self._HardwareThermocyclerMissingError as e:
            raise ThermocyclerNotOpenError(
                "Thermocycler must be open when moving to labware inside it,"
                " but can't confirm Thermocycler's current status."
            ) from e

        if (
            hw_tc_lid_status == ThermocyclerLidStatus.IN_BETWEEN
            or hw_tc_lid_status == ThermocyclerLidStatus.UNKNOWN
        ):
            # NOTE(cm): due to a potential hardware bug, the thermocycler might lose its position
            # status when idling in an open position and default to UNKNOWN or IN_BETWEEN.
            # This tries to recover only from an unexpected known position.
            await self._open_tc_lid(module_id=module_id)
            hw_tc_lid_status = await self._get_hardware_thermocycler_lid_status(
                module_id=module_id
            )
        if hw_tc_lid_status != ThermocyclerLidStatus.OPEN:
            raise ThermocyclerNotOpenError(
                f"Thermocycler must be open when moving to labware inside it,"
                f' but Thermocycler is currently "{hw_tc_lid_status}".'
            )

    async def _open_tc_lid(self, module_id: str) -> None:
        """Try to open the thermocycler lid."""
        tc_hardware = self._equipment.get_module_hardware_api(
            ThermocyclerModuleId(module_id)
        )
        if tc_hardware:
            await tc_hardware.open()

    async def ensure_labware_in_open_thermocycler(
        self, labware_parent: LabwareLocation
    ) -> None:
        """Flag unsafe movements to a Thermocycler.

        If the given labware is in a Thermocycler, and that Thermocycler's lid isn't
        currently open according the engine's thermocycler state as well as
        the hardware API (for non-virtual modules), tries to open the thermocycler lid.
        If this is unsuccessful, raises ThermocyclerNotOpenError.
        If it is a virtual module, checks only for thermocycler lid state in engine.

        Otherwise, no-ops.

        Warning:
            Various limitations with our hardware API and Thermocycler firmware mean
            this method can't detect every case where the Thermocycler lid is non-open.

            1. While the lid is in transit, the Thermocycler doesn't respond to status
               polls.
            2. The Thermocycler doesn't report when the user presses its physical
               lid open button; we have to wait for the next poll to find out, which
               can be tens of seconds because of (1).
            3. Nothing protects the Thermocycler hardware controller from being accessed
               concurrently by multiple tasks.
               And updating the Thermocycler state from one task doesn't guarantee that
               that update will be immediately visible from a different task.
               So, for example, if a legacy module control HTTP endpoint starts closing
               the Thermocycler, and then a Protocol Engine task quickly polls the
               Thermocycler through this method, this method may see the lid as open
               even though it's in transit.
        """
        if isinstance(labware_parent, ModuleLocation):
            module_id = labware_parent.moduleId
        else:
            return  # Labware not on a module.
        try:
            tc_substate = self._state_store.modules.get_thermocycler_module_substate(
                module_id=module_id
            )
        except WrongModuleTypeError:
            return  # Labware on a module, but not a Thermocycler.

        if not tc_substate.is_lid_open:
            raise ThermocyclerNotOpenError(
                "Thermocycler must be open when moving to labware inside it."
            )

        # There is a chance that the engine might not have the latest lid status;
        # do a hardware state check just to be sure that the lid is truly open.
        if not self._state_store.config.use_virtual_modules:
            await self._verify_tc_lid_status(module_id=module_id)

    async def _get_hardware_thermocycler_lid_status(
        self,
        module_id: str,
    ) -> ThermocyclerLidStatus:
        """Get lid status of the hardware Thermocycler corresponding with the module ID.

        Returns:
            Lid status of the requested attached thermocycler.

        Raises:
            _HardwareThermocyclerMissingError: If we can't find that Thermocycler in
                the hardware API, so we can't fetch its current lid status.
                It's unclear if this can happen in practice...
                maybe if the module disconnects between when it was loaded into
                Protocol Engine and when this function is called?
        """
        thermocycler_serial = self._state_store.modules.get_serial_number(
            module_id=module_id
        )
        thermocycler = await self._find_thermocycler_by_serial(
            serial_number=thermocycler_serial
        )
        if thermocycler is None:
            raise self._HardwareThermocyclerMissingError(
                f"No Thermocycler found" f' with serial number "{thermocycler_serial}".'
            )

        lid_status = thermocycler.lid_status
        # An attached thermocycler should always have lid status unless it's in error
        # or it was just connected and hasn't been polled for status yet.
        assert lid_status is not None, (
            "Did not receive a valid lid status from thermocycler. "
            "Cannot verify safe pipette movement"
        )
        return lid_status

    async def _find_thermocycler_by_serial(
        self, serial_number: str
    ) -> Optional[HardwareThermocycler]:
        """Find the hardware Thermocycler with the given serial number.

        Returns:
            The matching hardware Thermocycler, or None if none was found.
        """
        for module in self._hardware_api.attached_modules:
            # Different module types have different keys under .device_info.
            # Thermocyclers should always have .device_info["serial"].
            if (
                isinstance(module, HardwareThermocycler)
                and module.device_info["serial"] == serial_number
            ):
                return module
        return None

    class _HardwareThermocyclerMissingError(Exception):
        pass



"""Performance helpers for tracking robot activity."""

import functools
from pathlib import Path

from opentrons_shared_data.robot.types import RobotTypeEnum
import typing
from opentrons.config import (
    get_performance_metrics_data_dir,
    robot_configs,
    feature_flags as ff,
)

if typing.TYPE_CHECKING:
    from performance_metrics import RobotActivityState, SupportsTracking


_UnderlyingFunctionParameters = typing.ParamSpec("_UnderlyingFunctionParameters")
_UnderlyingFunctionReturn = typing.TypeVar("_UnderlyingFunctionReturn")
_UnderlyingFunction = typing.Callable[
    _UnderlyingFunctionParameters, _UnderlyingFunctionReturn
]


_should_track = ff.enable_performance_metrics(
    RobotTypeEnum.robot_literal_to_enum(robot_configs.load().model)
)


class _StubbedTracker:
    """A stubbed tracker that does nothing."""

    def __init__(self, storage_location: Path, should_track: bool) -> None:
        """Initialize the stubbed tracker."""
        pass

    def track(
        self,
        state: "RobotActivityState",
    ) -> typing.Callable[
        [_UnderlyingFunction[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn]],
        _UnderlyingFunction[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn],
    ]:
        """Return the original function."""

        def inner_decorator(
            func: _UnderlyingFunction[
                _UnderlyingFunctionParameters, _UnderlyingFunctionReturn
            ]
        ) -> _UnderlyingFunction[
            _UnderlyingFunctionParameters, _UnderlyingFunctionReturn
        ]:
            """Return the original function."""
            return func

        return inner_decorator

    def store(self) -> None:
        """Do nothing."""
        pass


# Ensure that _StubbedTracker implements SupportsTracking
# but do not create a runtime dependency on performance_metrics
if typing.TYPE_CHECKING:
    _: typing.Type["SupportsTracking"] = _StubbedTracker


def _handle_package_import() -> typing.Type["SupportsTracking"]:
    """Handle the import of the performance_metrics package.

    If the package is not available, return a stubbed tracker.
    """
    try:
        from performance_metrics import RobotActivityTracker

        return RobotActivityTracker
    except ImportError:
        return _StubbedTracker


_package_to_use = _handle_package_import()
_robot_activity_tracker: typing.Optional["SupportsTracking"] = None

# TODO: derek maggio (06-03-2024): investigate if _should_track should be
# reevaluated each time _get_robot_activity_tracker is called. I think this
# might get stuck in a state where after the first call, _should_track is
# always considered the initial value. It might miss changes to the feature
# flag. The easiest way to test this is on a robot when that is working.


def _get_robot_activity_tracker() -> "SupportsTracking":
    """Singleton for the robot activity tracker."""
    global _robot_activity_tracker
    if _robot_activity_tracker is None:
        _robot_activity_tracker = _package_to_use(
            get_performance_metrics_data_dir(), _should_track
        )
    return _robot_activity_tracker


def _track_a_function(
    state_name: "RobotActivityState",
    func: _UnderlyingFunction[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn],
) -> typing.Callable[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn]:
    """Wrap a passed function with RobotActivityTracker.track.

    This function is a decorator that will track the given state for the
    decorated function.

    Args:
        state_name: The state to annotate the tracked function with.
        func: The function to decorate.

    Returns:
        The decorated function.
    """
    tracker: SupportsTracking = _get_robot_activity_tracker()
    wrapped = tracker.track(state=state_name)(func)

    @functools.wraps(func)
    def wrapper(
        *args: _UnderlyingFunctionParameters.args,
        **kwargs: _UnderlyingFunctionParameters.kwargs
    ) -> _UnderlyingFunctionReturn:
        try:
            return wrapped(*args, **kwargs)
        finally:
            # TODO: derek maggio (06-18-2024): After investigation, it appears on startup
            # that the first call to tracker.store() will not actually store the data.
            # The second call stores both rows of data.

            tracker.store()

    return wrapper


class TrackingFunctions:
    """A class for tracking functions."""

    @staticmethod
    def track_analysis(
        func: _UnderlyingFunction[
            _UnderlyingFunctionParameters, _UnderlyingFunctionReturn
        ]
    ) -> typing.Callable[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn]:
        """Track a function that runs an analysis."""
        return _track_a_function("ANALYZING_PROTOCOL", func)

    @staticmethod
    def track_getting_cached_protocol_analysis(
        func: _UnderlyingFunction[
            _UnderlyingFunctionParameters, _UnderlyingFunctionReturn
        ]
    ) -> typing.Callable[_UnderlyingFunctionParameters, _UnderlyingFunctionReturn]:
        """Track a function that gets cached analysis."""
        return _track_a_function("GETTING_CACHED_ANALYSIS", func)



from abc import ABC, abstractmethod
from typing import Dict

from opentrons.drivers.types import Temperature


class AbstractTempDeckDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        ...

    @abstractmethod
    async def deactivate(self) -> None:
        ...

    @abstractmethod
    async def set_temperature(self, celsius: float) -> None:
        ...

    @abstractmethod
    async def get_temperature(self) -> Temperature:
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        """
        Queries Temp-Deck for its build version, model, and serial number

        returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Temp-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        ...

    @abstractmethod
    async def enter_programming_mode(self) -> None:
        ...



"""Protocol engine types to do with liquid handling."""

from dataclasses import dataclass
from typing import Dict


@dataclass
class FlowRates:
    """Default and current flow rates for a pipette."""

    default_blow_out: Dict[str, float]
    default_aspirate: Dict[str, float]
    default_dispense: Dict[str, float]



"""Getters for Protocol Engine motion planning."""
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from opentrons.types import Point
from opentrons.motion_planning.types import MoveType

from ..types import CurrentWell, CurrentPipetteLocation


@dataclass
class EdgeList:
    """Potential points for a touch tip operation."""

    right: Point
    left: Point
    center: Point
    forward: Point
    back: Point


class EdgePathType(str, Enum):
    """Types of well edge point paths for touch tip."""

    LEFT = "left"
    RIGHT = "right"
    DEFAULT = "default"


def get_move_type_to_well(
    pipette_id: str,
    labware_id: str,
    well_name: str,
    location: Optional[CurrentPipetteLocation],
    force_direct: bool,
) -> MoveType:
    """Get the move type for a move to well command."""
    if force_direct:
        return MoveType.DIRECT
    if (
        isinstance(location, CurrentWell)
        and pipette_id == location.pipette_id
        and labware_id == location.labware_id
    ):
        return (
            MoveType.IN_LABWARE_ARC
            if well_name != location.well_name
            else MoveType.DIRECT
        )
    else:
        return MoveType.GENERAL_ARC


def get_edge_point_list(
    center: Point,
    x_radius: float,
    y_radius: float,
    mm_from_edge: float,
    edge_path_type: EdgePathType,
) -> List[Point]:
    """Get list of edge points dependent on edge path type."""
    edges = EdgeList(
        right=center + Point(x=x_radius - mm_from_edge, y=0, z=0),
        left=center + Point(x=-x_radius + mm_from_edge, y=0, z=0),
        center=center,
        forward=center + Point(x=0, y=y_radius - mm_from_edge, z=0),
        back=center + Point(x=0, y=-y_radius + mm_from_edge, z=0),
    )

    if edge_path_type == EdgePathType.LEFT:
        return [edges.left, edges.center, edges.forward, edges.back, edges.center]
    elif edge_path_type == EdgePathType.RIGHT:
        return [edges.right, edges.center, edges.forward, edges.back, edges.center]
    else:
        return [
            edges.right,
            edges.left,
            edges.center,
            edges.forward,
            edges.back,
            edges.center,
        ]



"""Flex-specific extensions to instrument configuration."""
from typing import Union, Optional
from typing_extensions import Protocol

from .types import MountArgType

from opentrons.hardware_control.dev_types import (
    PipetteStateDict,
)
from opentrons.hardware_control.types import (
    TipStateType,
    InstrumentProbeType,
)
from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    PipetteOffsetSummary,
    GripperCalibrationOffset,
)


class FlexInstrumentConfigurer(Protocol[MountArgType]):
    """A protocol specifying Flex-specific extensions to instrument configuration."""

    async def get_instrument_state(
        self,
        mount: MountArgType,
    ) -> PipetteStateDict:
        ...

    def get_instrument_offset(
        self, mount: MountArgType
    ) -> Union[GripperCalibrationOffset, PipetteOffsetSummary, None]:
        ...

    async def get_tip_presence_status(
        self,
        mount: MountArgType,
    ) -> TipStateType:
        """Check tip presence status.

        If a high throughput pipette is present,
        move the tip motors down before checking the sensor status.
        """
        ...

    async def verify_tip_presence(
        self,
        mount: MountArgType,
        expected: TipStateType,
        follow_singular_sensor: Optional[InstrumentProbeType] = None,
    ) -> None:
        """Check tip presence status and raise if it does not match `expected`."""
        ...



"""Common resources for all advanced control functions."""
import enum
from typing import NamedTuple, Optional


class MixStrategy(enum.Enum):
    BOTH = enum.auto()
    BEFORE = enum.auto()
    AFTER = enum.auto()
    NEVER = enum.auto()


class MixOpts(NamedTuple):
    """
    Options to customize behavior of mix.

    These options will be passed to
    :py:meth:`InstrumentContext.mix` when it is called during the
    transfer.
    """

    repetitions: Optional[int] = None
    volume: Optional[float] = None
    rate: Optional[float] = None


MixOpts.repetitions.__doc__ = ":py:class:`int`"
MixOpts.volume.__doc__ = ":py:class:`float`"
MixOpts.rate.__doc__ = ":py:class:`float`"


class Mix(NamedTuple):
    """
    Options to control mix behavior before aspirate and after dispense.
    """

    mix_before: MixOpts = MixOpts()
    mix_after: MixOpts = MixOpts()



from .serial_connection import SerialConnection, AsyncResponseSerialConnection
from opentrons.drivers.asyncio.communication.errors import (
    SerialException,
    NoResponse,
    AlarmResponse,
    ErrorResponse,
    UnhandledGcode,
    DefaultErrorCodes,
)
from .async_serial import AsyncSerial

__all__ = [
    "SerialConnection",
    "AsyncResponseSerialConnection",
    "AsyncSerial",
    "SerialException",
    "NoResponse",
    "AlarmResponse",
    "ErrorResponse",
    "UnhandledGcode",
    "DefaultErrorCodes",
]



"""Command models to open the latch of a Flex Stacker."""

from __future__ import annotations

from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING
from typing_extensions import Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors import (
    ErrorOccurrence,
)

if TYPE_CHECKING:
    from ...state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

OpenLatchCommandType = Literal["flexStacker/openLatch"]


class OpenLatchParams(BaseModel):
    """The parameters defining how a stacker should open its latch."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")


class OpenLatchResult(BaseModel):
    """Result data from a stacker OpenLatch command."""


class OpenLatchImpl(AbstractCommandImpl[OpenLatchParams, SuccessData[OpenLatchResult]]):
    """Implementation of a stacker OpenLatch command."""

    def __init__(
        self, state_view: StateView, equipment: EquipmentHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: OpenLatchParams) -> SuccessData[OpenLatchResult]:
        """Execute the stacker OpenLatch command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        if stacker_hw is not None:
            await stacker_hw.open_latch()
        return SuccessData(public=OpenLatchResult())


class OpenLatch(BaseCommand[OpenLatchParams, OpenLatchResult, ErrorOccurrence]):
    """A command to OpenLatch the Flex Stacker of labware."""

    commandType: OpenLatchCommandType = "flexStacker/openLatch"
    params: OpenLatchParams
    result: Optional[OpenLatchResult] = None

    _ImplementationCls: Type[OpenLatchImpl] = OpenLatchImpl


class OpenLatchCreate(BaseCommandCreate[OpenLatchParams]):
    """A request to execute a Flex Stacker OpenLatch command."""

    commandType: OpenLatchCommandType = "flexStacker/openLatch"
    params: OpenLatchParams

    _CommandCls: Type[OpenLatch] = OpenLatch



"""Magnetic block sub-state."""

from dataclasses import dataclass
from typing import NewType

MagneticBlockId = NewType("MagneticBlockId", str)


@dataclass(frozen=True)
class MagneticBlockSubState:
    """Magnetic Block specific state.

    Provides a read-only state access
    for an individual loaded Magnetic Block.
    """

    module_id: MagneticBlockId



"""Mappings between API versions and overlap versions."""
from functools import lru_cache
from typing_extensions import Final
from opentrons.protocols.api_support.types import APIVersion

_OVERLAP_VERSION_MAP: Final = {
    APIVersion(2, 0): "v0",
    APIVersion(2, 19): "v1",
    APIVersion(2, 20): "v3",
}


@lru_cache(1)
def overlap_for_api_version(api_version: APIVersion) -> str:
    """Get the overlap version for a specific API version."""
    defined = list(reversed(sorted(_OVERLAP_VERSION_MAP.keys())))
    for version in defined:
        if version <= api_version:
            return _OVERLAP_VERSION_MAP[version]
    return _OVERLAP_VERSION_MAP[APIVersion(2, 0)]



"""Core protocol logic interfaces and implementations.

This module provide facades to different protocol execution cores,
and is an internal implementation detail of the Python Protocol API.

Nothing defined in `opentrons.protocol_api.core` may be considered public.
"""
# TODO(mc, 2022-08-22): uncomment when import cycles can be resolved
# from .protocol import AbstractProtocol
# from .labware_offset_provider import (
#     AbstractLabwareOffsetProvider,
#     LabwareOffsetProvider,
#     NullLabwareOffsetProvider,
#     ProvidedLabwareOffset,
# )

# __all__ = [
#     "AbstractProtocol",
#     "AbstractLabwareOffsetProvider",
#     "LabwareOffsetProvider",
#     "NullLabwareOffsetProvider",
#     "ProvidedLabwareOffset",
# ]



"""Protocol engine types to do with deck configuration."""

from dataclasses import dataclass
from typing import FrozenSet, List, Tuple, Optional
from enum import Enum

from opentrons.types import DeckSlotName

from opentrons_shared_data.module.types import ModuleType as SharedDataModuleType

from .util import Vec3f, Dimensions


class AddressableOffsetVector(Vec3f):
    """Offset, in deck coordinates, from nominal to actual position of an addressable area."""


@dataclass(frozen=True)
class PotentialCutoutFixture:
    """Cutout and cutout fixture id associated with a potential cutout fixture that can be on the deck."""

    cutout_id: str
    cutout_fixture_id: str
    provided_addressable_areas: FrozenSet[str]


class AreaType(Enum):
    """The type of addressable area."""

    SLOT = "slot"
    STAGING_SLOT = "stagingSlot"
    MOVABLE_TRASH = "movableTrash"
    FIXED_TRASH = "fixedTrash"
    WASTE_CHUTE = "wasteChute"
    THERMOCYCLER = "thermocycler"
    HEATER_SHAKER = "heaterShaker"
    TEMPERATURE = "temperatureModule"
    MAGNETICBLOCK = "magneticBlock"
    ABSORBANCE_READER = "absorbanceReader"
    FLEX_STACKER = "flexStacker"
    LID_DOCK = "lidDock"


@dataclass(frozen=True)
class AddressableArea:
    """Addressable area that has been loaded."""

    area_name: str
    area_type: AreaType
    base_slot: DeckSlotName
    display_name: str
    bounding_box: Dimensions
    position: AddressableOffsetVector
    compatible_module_types: List[SharedDataModuleType]


# TODO make the below some sort of better type
# TODO This should instead contain a proper cutout fixture type
DeckConfigurationType = List[
    Tuple[str, str, Optional[str]]
]  # cutout_id, cutout_fixture_id, opentrons_module_serial_number


# TODO(mm, 2023-05-10): Deduplicate with constants in
# opentrons.protocols.api_support.deck_type
# and consider moving to shared-data.
class DeckType(str, Enum):
    """Types of deck available."""

    OT2_STANDARD = "ot2_standard"
    OT2_SHORT_TRASH = "ot2_short_trash"
    OT3_STANDARD = "ot3_standard"



from abc import abstractmethod, ABC
from typing import Optional, Union

from opentrons.types import AxisMapType, Mount, Point
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons.protocol_api._types import PlungerPositionTypes, PipetteActionTypes


class AbstractRobot(ABC):
    @abstractmethod
    def get_pipette_type_from_engine(
        self, mount: Union[Mount, str]
    ) -> Optional[PipetteNameType]:
        ...

    @abstractmethod
    def get_plunger_position_from_volume(
        self, mount: Mount, volume: float, action: PipetteActionTypes, robot_type: str
    ) -> float:
        ...

    @abstractmethod
    def get_plunger_position_from_name(
        self, mount: Mount, position_name: PlungerPositionTypes
    ) -> float:
        ...

    @abstractmethod
    def move_to(self, mount: Mount, destination: Point, speed: Optional[float]) -> None:
        ...

    @abstractmethod
    def move_axes_to(
        self,
        axis_map: AxisMapType,
        critical_point: Optional[AxisMapType],
        speed: Optional[float],
    ) -> None:
        ...

    @abstractmethod
    def move_axes_relative(self, axis_map: AxisMapType, speed: Optional[float]) -> None:
        ...

    @abstractmethod
    def release_grip(self) -> None:
        ...

    @abstractmethod
    def close_gripper(self, force: Optional[float] = None) -> None:
        ...



from typing import Dict, List, Optional, Mapping
from typing_extensions import Protocol

from opentrons.types import Point
from ..types import Axis, CriticalPoint, MotionChecks
from .types import MountArgType


class MotionController(Protocol[MountArgType]):
    """Protocol specifying fundamental motion controls."""

    def get_deck_from_machine(
        self, machine_pos: Dict[Axis, float]
    ) -> Dict[Axis, float]:
        """Convert machine coordinates to deck coordinates."""
        ...

    async def halt(self, disengage_before_stopping: bool = False) -> None:
        """Immediately stop motion.

        Calls to stop through the synch adapter while other calls
        are ongoing will typically wait until those calls are done, since most
        of the async calls here in fact block the loop while they talk to
        smoothie. To provide actual immediate halting, call this method which
        does not require use of the loop.

        If disengage_before_stopping is True, the motors will disengage first and then
        stop in place. Disengaging creates a smoother halt but requires homing after
        in order to resume movement.
        """
        ...

    async def stop(self, home_after: bool = True) -> None:
        """
        Stop motion as soon as possible, reset, and optionally home.

        This will cancel motion (after the current call to :py:meth:`move`;
        see :py:meth:`pause` for more detail), then home and reset the
        robot.
        """
        ...

    async def reset(self) -> None:
        """Reset the stored state of the system.

        This will re-scan instruments and models, clearing any cached
        information about their presence or state.
        """
        ...

    # Gantry/frame (i.e. not pipette) action API
    async def home_z(
        self,
        mount: Optional[MountArgType] = None,
        allow_home_other: bool = True,
    ) -> None:
        """Home a selected z-axis, or both if not specified."""
        ...

    async def home_plunger(self, mount: MountArgType) -> None:
        """
        Home the plunger motor for a mount, and then return it to the 'bottom'
        position.

        mount: the mount associated with the target plunger
        """
        ...

    async def home(self, axes: Optional[List[Axis]] = None) -> None:
        """Home a list of axes and initialize current position.

        axes A list of axes to home. Default is `None`, which will
             home everything.
        """
        ...

    async def current_position(
        self,
        mount: MountArgType,
        critical_point: Optional[CriticalPoint] = None,
        refresh: bool = False,
        # TODO(mc, 2021-11-15): combine with `refresh` for more reliable
        # position reporting when motors are not homed
        fail_on_not_homed: bool = False,
    ) -> Dict[Axis, float]:
        """Return the postion (in deck coords) of the critical point of the
        specified mount.

        This returns cached position to avoid hitting the smoothie driver
        unless ``refresh`` is ``True``.

        If `critical_point` is specified, that critical point will be applied
        instead of the default one. For instance, if
        `critical_point=CriticalPoints.MOUNT` then the position of the mount
        will be returned. If the critical point specified does not exist, then
        the next one down is returned - for instance, if there is no tip on the
        specified mount but `CriticalPoint.TIP` was specified, the position of
        the nozzle will be returned.

        If `fail_on_not_homed` is `True`, this method will raise a `PositionUnknownError`
        if any of the relavent axes are not homed, regardless of `refresh`.
        """
        ...

    async def gantry_position(
        self,
        mount: MountArgType,
        critical_point: Optional[CriticalPoint] = None,
        refresh: bool = False,
        # TODO(mc, 2021-11-15): combine with `refresh` for more reliable
        # position reporting when motors are not homed
        fail_on_not_homed: bool = False,
    ) -> Point:
        """Retrieve the position of just the currently-active pipette.

        While current_position returns the position of every actuator on the system,
        this function just returns the x, y, and z of the critical point of whichever
        pipette is currently active (last moved).
        """
        ...

    async def move_to(
        self,
        mount: MountArgType,
        abs_position: Point,
        speed: Optional[float] = None,
        critical_point: Optional[CriticalPoint] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
    ) -> None:
        """Move the critical point of the specified mount to a location
        relative to the deck, at the specified speed. 'speed' sets the speed
        of all robot axes to the given value. So, if multiple axes are to be
        moved, they will do so at the same speed

        The critical point of the mount depends on the current status of
        the mount:
        - If the mount does not have anything attached, its critical point is
          the bottom of the mount attach bracket.
        - If the mount has a pipette attached and it is not known to have a
          pipette tip, the critical point is the end of the nozzle of a single
          pipette or the end of the backmost nozzle of a multipipette
        - If the mount has a pipette attached and it is known to have a
          pipette tip, the critical point is the end of the pipette tip for
          a single pipette or the end of the tip of the backmost nozzle of a
          multipipette

        :param mount: The mount to move
        :param abs_position: The target absolute position in
                             deck coordinates to move the
                             critical point to
        :param speed: An overall head speed to use during the move
        :param critical_point: The critical point to move. In most situations
                               this is not needed. If not specified, the
                               current critical point will be moved. If
                               specified, the critical point must be one that
                               actually exists - that is, specifying
                               :py:attr:`.CriticalPoint.NOZZLE` when no pipette
                               is attached or :py:attr:`.CriticalPoint.TIP`
                               when no tip is applied will result in an error.
        :param max_speeds: An optional override for per-axis maximum speeds. If
                           an axis is specified, it will not move faster than
                           the given speed. Note that this does not make that
                           axis move precisely at the given speed; it only
                           it if it was going to go faster. Direct speed
                           is still set by ``speed``.
        """
        ...

    async def move_axes(
        self,
        position: Mapping[Axis, float],
        speed: Optional[float] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
        expect_stalls: bool = False,
    ) -> None:
        """Moves the effectors of the specified axis to the specified position.
        The effector of the x,y axis is the center of the carriage.
        The effector of the pipette mount axis are the mount critical points but only in z.
        """
        ...

    async def move_rel(
        self,
        mount: MountArgType,
        delta: Point,
        speed: Optional[float] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
        check_bounds: MotionChecks = MotionChecks.NONE,
        fail_on_not_homed: bool = False,
    ) -> None:
        """Move the critical point of the specified mount by a specified
        displacement in a specified direction, at the specified speed.
        'speed' sets the speed of all axes to the given value. So, if multiple
        axes are to be moved, they will do so at the same speed.

        If fail_on_not_homed is True (default False), if an axis that is not
        homed moves it will raise a PositionUnknownError. Otherwise, it will home the axis.
        """
        ...

    def get_engaged_axes(self) -> Dict[Axis, bool]:
        """Which axes are engaged and holding."""
        ...

    @property
    def engaged_axes(self) -> Dict[Axis, bool]:
        """Which axes are engaged and holding"""
        return self.get_engaged_axes()

    async def disengage_axes(self, which: List[Axis]) -> None:
        """Disengage some axes."""
        ...

    async def engage_axes(self, which: List[Axis]) -> None:
        """Engage some axes."""
        ...

    async def retract(self, mount: MountArgType, margin: float = 10) -> None:
        """Pull the specified mount up to its home position.

        Works regardless of critical point or home status.
        """
        ...

    async def retract_axis(self, axis: Axis) -> None:
        """Retract the specified axis to its home position."""
        ...

    def is_movement_execution_taskified(self) -> bool:
        """Get whether move functions are being executed inside cancellable tasks."""
        ...

    def should_taskify_movement_execution(self, taskify: bool) -> None:
        """Specify whether move functions should be executed inside cancellable tasks."""
        ...

    async def cancel_execution_and_running_tasks(self) -> None:
        """Cancel all tasks and set execution manager state to Cancelled."""
        ...

    async def prepare_for_mount_movement(self, mount: MountArgType) -> None:
        """Retract the other mount if necessary."""
        ...



"""Common flex stacker base models."""
from typing import Literal

from ...errors import ErrorOccurrence
from opentrons_shared_data.errors import ErrorCodes


class FlexStackerStallOrCollisionError(ErrorOccurrence):
    """Returned when the motor driver detects a stall."""

    isDefined: bool = True
    errorType: Literal["flexStackerStallOrCollision"] = "flexStackerStallOrCollision"

    errorCode: str = ErrorCodes.FLEX_STACKER_STALL_OR_COLLISION_DETECTED.value.code
    detail: str = ErrorCodes.FLEX_STACKER_STALL_OR_COLLISION_DETECTED.value.detail



import asyncio
import logging
import pathlib
import time
from typing import Callable, Dict, Tuple
from opentrons.hardware_control.types import BoardRevision, DoorState
from . import RevisionPinsError
from .types import gpio_group, PinDir, GPIOPin

import gpiod  # type: ignore[import-not-found]

"""
Raspberry Pi GPIO control module
"""

MODULE_LOG = logging.getLogger(__name__)

DTOVERLAY_PATH = "/proc/device-tree/soc/gpio@7e200000/gpio_rev_bit_pins"


def _event_callback(
    update_door_state: Callable[[DoorState], None],
    get_door_state: Callable[..., DoorState],
) -> None:
    try:
        update_door_state(get_door_state())
    except Exception:
        MODULE_LOG.exception("Errored during event callback")


class GPIOCharDev:
    def __init__(self, chip_name: str):
        self._board_rev = BoardRevision.UNKNOWN
        self._chip = gpiod.Chip(chip_name)
        self._lines = self._initialize()

    @property
    def chip(self) -> gpiod.Chip:
        return self._chip

    @property
    def lines(self) -> Dict[str, gpiod.Line]:
        return self._lines

    @property
    def board_rev(self) -> BoardRevision:
        return self._board_rev

    @board_rev.setter
    def board_rev(self, boardrev: BoardRevision) -> None:
        self._board_rev = boardrev

    def _request_line(self, pin: GPIOPin, request_type: int) -> gpiod.Line:
        name = pin.name
        offset = pin.by_board_rev(self.board_rev)

        line = self.chip.get_line(offset)

        def _retry_request_line(retries: int = 0) -> gpiod.Line:
            try:
                line.request(consumer=name, type=request_type, default_vals=[0])
            except OSError as e:
                retries -= 1
                if retries <= 0:
                    raise e
                time.sleep(0.25)
                return _retry_request_line(retries)
            return line

        if name == "blue_button":
            # todo(mm, 2021-10-22): See if we can remove these retries.
            # They may have been to paper over a former race condition where, during
            # boot, the robot server would start initializing hardware before the
            # standalone light-blinking process ot-blinkenlights was guaranteed to
            # have relinquished its handle on the blue button GPIO line.
            return _retry_request_line(3)
        else:
            return _retry_request_line()

    def _initialize(self) -> Dict[str, gpiod.Line]:
        MODULE_LOG.info("Registering Central Routing Board Revision GPIOs")
        lines = {}
        rev_pins = gpio_group.by_names(["rev_0", "rev_1"])
        for rev in rev_pins.pins:
            lines[rev.name] = self._request_line(rev, gpiod.LINE_REQ_DIR_IN)
        return lines

    def _determine_board_revision(self) -> BoardRevision:
        """Read revision bit pins and return the board revision"""
        try:
            rev_bits = self.read_revision_bits()
            return BoardRevision.by_bits(rev_bits)
        except RevisionPinsError:
            MODULE_LOG.info(
                "Failed to detect central routing board revision gpio "
                "pins, defaulting to 2.1 (OG)"
            )
            return BoardRevision.OG
        except Exception:
            MODULE_LOG.exception(
                "Unexpected error from reading central routing board " "revision bits"
            )
            return BoardRevision.UNKNOWN

    def config_by_board_rev(self) -> None:
        MODULE_LOG.info("Configuring GPIOs by central routing roard revision")
        # get board revision based on rev bits
        self.board_rev = self._determine_board_revision()

        # setup output lines
        output_pins = gpio_group.by_type(PinDir.output)
        for output in output_pins.pins:
            self._lines[output.name] = self._request_line(
                output, gpiod.LINE_REQ_DIR_OUT
            )

        # setup input lines
        input_pins = gpio_group.by_type(PinDir.input)
        sorted_inputs = input_pins.group_by_pins(self.board_rev)
        for pins in sorted_inputs:
            line = self._request_line(pins[0], gpiod.LINE_REQ_EV_BOTH_EDGES)
            for pin in pins:
                self._lines[pin.name] = line

    async def setup(self) -> None:
        MODULE_LOG.info("Setting up GPIOs")
        self.set_audio_enable(True)
        # smoothieware programming pins, must be in a known state (HIGH)
        self.set_halt_pin(True)
        self.set_isp_pin(True)
        self.set_reset_pin(False)
        await asyncio.sleep(0.25)
        self.set_reset_pin(True)
        await asyncio.sleep(0.25)

    def setup_v1(self) -> None:
        # TODO: AA 07-08-2020 remove when legacy API is deprecated
        MODULE_LOG.info("Setting up GPIOs for APIv1")
        self.set_audio_enable(True)
        # smoothieware programming pins, must be in a known state (HIGH)
        self.set_halt_pin(True)
        self.set_isp_pin(True)
        self.set_reset_pin(False)
        time.sleep(0.25)
        self.set_reset_pin(True)
        time.sleep(0.25)

    def set_high(self, output_pin: GPIOPin) -> None:
        self.lines[output_pin.name].set_value(1)

    def set_low(self, output_pin: GPIOPin) -> None:
        self.lines[output_pin.name].set_value(0)

    def set_button_light(
        self, red: bool = False, green: bool = False, blue: bool = False
    ) -> None:
        color_pins = {
            gpio_group.red_button: red,
            gpio_group.green_button: green,
            gpio_group.blue_button: blue,
        }
        for pin, state in color_pins.items():
            if state:
                self.set_high(pin)
            else:
                self.set_low(pin)

    def set_rail_lights(self, on: bool = True) -> None:
        if on:
            self.set_high(gpio_group.frame_leds)
        else:
            self.set_low(gpio_group.frame_leds)

    def set_reset_pin(self, on: bool = True) -> None:
        if on:
            self.set_high(gpio_group.reset)
        else:
            self.set_low(gpio_group.reset)

    def set_isp_pin(self, on: bool = True) -> None:
        if on:
            self.set_high(gpio_group.isp)
        else:
            self.set_low(gpio_group.isp)

    def set_halt_pin(self, on: bool = True) -> None:
        if on:
            self.set_high(gpio_group.halt)
        else:
            self.set_low(gpio_group.halt)

    def set_audio_enable(self, on: bool = True) -> None:
        if on:
            self.set_high(gpio_group.audio_enable)
        else:
            self.set_low(gpio_group.audio_enable)

    def _read(self, input_pin: GPIOPin) -> gpiod.Line:
        try:
            return self.lines[input_pin.name].get_value()
        except KeyError:
            raise RuntimeError(
                f"GPIO {input_pin.name} is not registered and cannot" "be read"
            )

    def get_button_light(self) -> Tuple[bool, bool, bool]:
        return (
            bool(self._read(gpio_group.red_button)),
            bool(self._read(gpio_group.green_button)),
            bool(self._read(gpio_group.blue_button)),
        )

    def get_rail_lights(self) -> bool:
        return bool(self._read(gpio_group.frame_leds))

    def read_button(self) -> bool:
        # button is normal-HIGH, so invert
        return not bool(self._read(gpio_group.button_input))

    def read_window_switches(self) -> bool:
        return bool(self._read(gpio_group.window_door_sw))

    def read_top_window_switch(self) -> bool:
        return bool(self._read(gpio_group.window_sw_filt))

    def read_front_door_switch(self) -> bool:
        return bool(self._read(gpio_group.door_sw_filt))

    def read_revision_bits(self) -> Tuple[bool, bool]:
        """Read revision bit GPIO pins

        If the gpio_rev_bit_pins device tree overlay is enabled,
        returns the pins' values. Otherwise, return RevisionPinsError
        """
        if pathlib.Path(DTOVERLAY_PATH).exists():
            return (
                bool(self._read(gpio_group.rev_0)),
                bool(self._read(gpio_group.rev_1)),
            )
        else:
            raise RevisionPinsError

    def get_door_state(self) -> DoorState:
        name = gpio_group.window_door_sw.name
        event = self.lines[name].event_read()
        if event.type == gpiod.LineEvent.FALLING_EDGE:
            return DoorState.OPEN
        else:
            return DoorState.CLOSED

    def start_door_switch_watcher(
        self,
        loop: asyncio.AbstractEventLoop,
        update_door_state: Callable[[DoorState], None],
    ) -> None:
        current_door_value = self.read_window_switches()
        if current_door_value == 0:
            update_door_state(DoorState.OPEN)
        else:
            update_door_state(DoorState.CLOSED)

        try:
            door_fd = self.lines["window_door_sw"].event_get_fd()
            loop.add_reader(
                door_fd, _event_callback, update_door_state, self.get_door_state
            )
        except Exception:
            MODULE_LOG.exception(
                "Failed to add fd reader, cannot not monitor window door "
                "switch properly"
            )

    def release_line(self, pin: GPIOPin) -> None:
        self.lines[pin.name].release()
        self.lines.pop(pin.name)

    def stop_door_switch_watcher(self, loop: asyncio.AbstractEventLoop) -> None:
        try:
            door_fd = self.lines["window_door_sw"].event_get_fd()
            loop.remove_reader(door_fd)
        except Exception:
            MODULE_LOG.exception("Failed to remove window door switch fd reader")



"""Command models to configure the stored labware pool of a Flex Stacker.."""

from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING, Annotated
from typing_extensions import Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors import (
    ErrorOccurrence,
)
from ...errors.exceptions import FlexStackerNotLogicallyEmptyError
from ...state import update_types

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

SetStoredLabwareCommandType = Literal["flexStacker/setStoredLabware"]


class StackerStoredLabwareDetails(BaseModel):
    """The parameters defining a labware to be stored in the stacker."""

    loadName: str = Field(
        ..., description="Name used to reference the definition of this labware."
    )
    namespace: str = Field(
        ..., description="Namespace of the definition of this labware."
    )
    version: int = Field(..., description="Version of the definition of this labware.")


class SetStoredLabwareParams(BaseModel):
    """Input parameters for a setStoredLabware command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the Flex Stacker.",
    )
    primaryLabware: StackerStoredLabwareDetails = Field(
        ...,
        description="The details of the primary labware (i.e. not the lid or adapter, if any) stored in the stacker.",
    )
    lidLabware: StackerStoredLabwareDetails | SkipJsonSchema[None] = Field(
        default=None,
        description="The details of the lid on the primary labware, if any.",
    )
    adapterLabware: StackerStoredLabwareDetails | SkipJsonSchema[None] = Field(
        default=None,
        description="The details of the adapter under the primary labware, if any.",
    )
    initialCount: Optional[Annotated[int, Field(ge=0)]] = Field(
        None,
        description=(
            "The number of labware that should be initially stored in the stacker. This number will be silently clamped to "
            "the maximum number of labware that will fit; do not rely on the parameter to know how many labware are in the stacker."
        ),
    )


class SetStoredLabwareResult(BaseModel):
    """Result data from a setstoredlabware command."""

    primaryLabwareDefinition: LabwareDefinition = Field(
        ..., description="The definition of the primary labware."
    )
    lidLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = Field(
        None, description="The definition of the lid on the primary labware, if any."
    )
    adapterLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = Field(
        None,
        description="The definition of the adapter under the primary labware, if any.",
    )
    count: int = Field(
        ..., description="The number of labware loaded into the stacker labware pool."
    )


class SetStoredLabwareImpl(
    AbstractCommandImpl[SetStoredLabwareParams, SuccessData[SetStoredLabwareResult]]
):
    """Implementation of a setstoredlabware command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: SetStoredLabwareParams
    ) -> SuccessData[SetStoredLabwareResult]:
        """Execute the setstoredlabwarecommand."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        if stacker_state.pool_count != 0:
            # Note: this error catches if the protocol tells us the stacker is not empty, making this command
            # invalid at this point in the protocol. This error is not recoverable and should occur during
            # analysis; the protocol must be changed.

            raise FlexStackerNotLogicallyEmptyError(
                message=(
                    "The Flex Stacker must be known to be empty before reconfiguring its labware pool, but it has "
                    f"a pool of {stacker_state.pool_count} labware"
                )
            )

        labware_def, _ = await self._equipment.load_definition_for_details(
            load_name=params.primaryLabware.loadName,
            namespace=params.primaryLabware.namespace,
            version=params.primaryLabware.version,
        )
        lid_def: LabwareDefinition | None = None
        if params.lidLabware:
            lid_def, _ = await self._equipment.load_definition_for_details(
                load_name=params.lidLabware.loadName,
                namespace=params.lidLabware.namespace,
                version=params.lidLabware.version,
            )
        adapter_def: LabwareDefinition | None = None
        if params.adapterLabware:
            adapter_def, _ = await self._equipment.load_definition_for_details(
                load_name=params.adapterLabware.loadName,
                namespace=params.adapterLabware.namespace,
                version=params.adapterLabware.version,
            )

        self._state_view.labware.raise_if_stacker_labware_pool_is_not_valid(
            labware_def, lid_def, adapter_def
        )

        pool_height = self._state_view.geometry.get_height_of_labware_stack(
            [x for x in [lid_def, labware_def, adapter_def] if x is not None]
        )
        max_pool_count = self._state_view.modules.stacker_max_pool_count_by_height(
            params.moduleId, pool_height
        )

        initial_count = (
            params.initialCount if params.initialCount is not None else max_pool_count
        )
        count = min(initial_count, max_pool_count)

        state_update = (
            update_types.StateUpdate()
            .update_flex_stacker_labware_pool_definition(
                params.moduleId, max_pool_count, labware_def, adapter_def, lid_def
            )
            .update_flex_stacker_labware_pool_count(params.moduleId, count)
        )
        return SuccessData(
            public=SetStoredLabwareResult.model_construct(
                primaryLabwareDefinition=labware_def,
                lidLabwareDefinition=lid_def,
                adapterLabwareDefinition=adapter_def,
                count=count,
            ),
            state_update=state_update,
        )


class SetStoredLabware(
    BaseCommand[SetStoredLabwareParams, SetStoredLabwareResult, ErrorOccurrence]
):
    """A command to setstoredlabware the Flex Stacker."""

    commandType: SetStoredLabwareCommandType = "flexStacker/setStoredLabware"
    params: SetStoredLabwareParams
    result: Optional[SetStoredLabwareResult] = None

    _ImplementationCls: Type[SetStoredLabwareImpl] = SetStoredLabwareImpl


class SetStoredLabwareCreate(BaseCommandCreate[SetStoredLabwareParams]):
    """A request to execute a Flex Stacker SetStoredLabware command."""

    commandType: SetStoredLabwareCommandType = "flexStacker/setStoredLabware"
    params: SetStoredLabwareParams

    _CommandCls: Type[SetStoredLabware] = SetStoredLabware



"""Command models to deactivate a Temperature Module."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

DeactivateTemperatureCommandType = Literal["temperatureModule/deactivate"]


class DeactivateTemperatureParams(BaseModel):
    """Input parameters to deactivate a Temperature Module."""

    moduleId: str = Field(..., description="Unique ID of the Temperature Module.")


class DeactivateTemperatureResult(BaseModel):
    """Result data from deactivating a Temperature Module."""


class DeactivateTemperatureImpl(
    AbstractCommandImpl[
        DeactivateTemperatureParams, SuccessData[DeactivateTemperatureResult]
    ]
):
    """Execution implementation of a Temperature Module's deactivate command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: DeactivateTemperatureParams
    ) -> SuccessData[DeactivateTemperatureResult]:
        """Deactivate a Temperature Module."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        module_substate = self._state_view.modules.get_temperature_module_substate(
            module_id=params.moduleId
        )

        # Allow propagation of ModuleNotAttachedError.
        temp_hardware_module = self._equipment.get_module_hardware_api(
            module_substate.module_id
        )

        if temp_hardware_module is not None:
            await temp_hardware_module.deactivate()
        return SuccessData(
            public=DeactivateTemperatureResult(),
        )


class DeactivateTemperature(
    BaseCommand[
        DeactivateTemperatureParams, DeactivateTemperatureResult, ErrorOccurrence
    ]
):
    """A command to deactivate a Temperature Module."""

    commandType: DeactivateTemperatureCommandType = "temperatureModule/deactivate"

    params: DeactivateTemperatureParams
    result: Optional[DeactivateTemperatureResult] = None

    _ImplementationCls: Type[DeactivateTemperatureImpl] = DeactivateTemperatureImpl


class DeactivateTemperatureCreate(BaseCommandCreate[DeactivateTemperatureParams]):
    """A request to deactivate a Temperature Module."""

    commandType: DeactivateTemperatureCommandType = "temperatureModule/deactivate"
    params: DeactivateTemperatureParams

    _CommandCls: Type[DeactivateTemperature] = DeactivateTemperature



"""ProtocolEngine class definition."""

from contextlib import AsyncExitStack
from logging import getLogger
from typing import Dict, Optional, Union, AsyncGenerator, Callable

from opentrons_shared_data.errors import (
    ErrorCodes,
    EnumeratedError,
)
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules import AbstractModule as HardwareModuleAPI
from opentrons.hardware_control.types import PauseType as HardwarePauseType

from .actions.actions import (
    ResumeFromRecoveryAction,
    SetErrorRecoveryPolicyAction,
)
from .errors import ProtocolCommandFailedError, ErrorOccurrence, CommandNotAllowedError
from .errors.exceptions import EStopActivatedError
from .error_recovery_policy import ErrorRecoveryPolicy
from . import commands, slot_standardization, labware_offset_standardization
from .resources import ModelUtils, ModuleDataProvider, FileProvider
from .types import (
    LabwareOffset,
    LabwareOffsetCreate,
    LegacyLabwareOffsetCreate,
    LabwareUri,
    ModuleModel,
    Liquid,
    HexColor,
    PostRunHardwareState,
    DeckConfigurationType,
)
from .execution import (
    QueueWorker,
    create_queue_worker,
    DoorWatcher,
    HardwareStopper,
)
from .state.state import StateStore, StateView
from .state.update_types import StateUpdate
from .plugins import AbstractPlugin, PluginStarter
from .actions import (
    ActionDispatcher,
    PlayAction,
    PauseAction,
    PauseSource,
    StopAction,
    FinishAction,
    FinishErrorDetails,
    QueueCommandAction,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    AddLiquidAction,
    SetDeckConfigurationAction,
    AddAddressableAreaAction,
    AddModuleAction,
    HardwareStoppedAction,
    ResetTipsAction,
    SetPipetteMovementSpeedAction,
)


_log = getLogger(__name__)


class ProtocolEngine:
    """Main ProtocolEngine class.

    A ProtocolEngine instance holds the state of a protocol as it executes,
    and manages calls to a command executor that actually implements the logic
    of the commands themselves.

    Lifetime:
        Instances are single-use. Each instance is associated with a single protocol,
        or a a single chain of robot control such as Labware Position Check.

    Concurrency:
        Instances live in `asyncio` event loops. Each instance must be constructed inside an
        event loop, and then must be interacted with exclusively through that
        event loop's thread--even for regular non-`async` methods, like `.pause()`.
        (This is because there are background async tasks that monitor state changes using
        primitives that aren't thread-safe. See ChangeNotifier.)
    """

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
        plugin_starter: PluginStarter,
        model_utils: ModelUtils,
        hardware_stopper: HardwareStopper,
        door_watcher: DoorWatcher,
        module_data_provider: ModuleDataProvider,
        file_provider: FileProvider,
        queue_worker: Optional[QueueWorker] = None,
    ) -> None:
        """Initialize a ProtocolEngine instance.

        Must be called while an event loop is active.

        This constructor is only for `ProtocolEngine` unit tests.
        Prefer the `create_protocol_engine()` factory function.
        """
        self._hardware_api = hardware_api
        self._file_provider = file_provider
        self._state_store = state_store
        self._model_utils = model_utils
        self._action_dispatcher = action_dispatcher
        self._plugin_starter = plugin_starter
        self._hardware_stopper = hardware_stopper
        self._door_watcher = door_watcher
        self._module_data_provider = module_data_provider
        self._queue_worker = queue_worker
        if self._queue_worker:
            self._queue_worker.start()
        self._door_watcher.start()

    @property
    def state_view(self) -> StateView:
        """Get an interface to retrieve calculated state values."""
        return self._state_store

    @property
    def _get_queue_worker(self) -> QueueWorker:
        """Get the queue worker instance."""
        assert self._queue_worker is not None
        return self._queue_worker

    def add_plugin(self, plugin: AbstractPlugin) -> None:
        """Add a plugin to the engine to customize behavior."""
        self._plugin_starter.start(plugin)

    def set_deck_configuration(
        self, deck_configuration: Optional[DeckConfigurationType]
    ) -> None:
        """Inform the engine of the robot's current deck configuration.

        If `Config.use_simulated_deck_config` is `True`, this is meaningless and unused.
        You can call this with `None` if you want to be explicit--it will no-op.

        If `Config.use_simulated_deck_config` is `False`, you should call this with the
        robot's actual, full, non-`None` deck configuration, before you play the run for
        the first time. Do not call this in the middle of a run.
        """
        self._action_dispatcher.dispatch(SetDeckConfigurationAction(deck_configuration))

    def play(self) -> None:
        """Start or resume executing commands in the queue."""
        requested_at = self._model_utils.get_timestamp()
        # TODO(mc, 2021-08-05): if starting, ensure plungers motors are
        # homed if necessary
        action = self._state_store.commands.validate_action_allowed(
            PlayAction(requested_at=requested_at)
        )
        self._action_dispatcher.dispatch(action)

        if self._state_store.commands.get_is_door_blocking():
            self._hardware_api.pause(HardwarePauseType.PAUSE)
        else:
            self._hardware_api.resume(HardwarePauseType.PAUSE)

    def request_pause(self) -> None:
        """Make command execution pause soon.

        This will try to pause in the middle of the ongoing command, if there is one.
        Otherwise, whenever the next command begins, the pause will happen then.
        """
        action = self._state_store.commands.validate_action_allowed(
            PauseAction(source=PauseSource.CLIENT)
        )
        self._action_dispatcher.dispatch(action)
        self._hardware_api.pause(HardwarePauseType.PAUSE)

    def resume_from_recovery(self, reconcile_false_positive: bool) -> None:
        """Resume normal protocol execution after the engine was `AWAITING_RECOVERY`.

        If `reconcile_false_positive` is `False`, the engine will continue naively from
        whatever state the error left it in. (Each defined error individually documents
        exactly how it affects state.) This is appropriate for client-driven error
        recovery, where the client wants predictable behavior from the engine.

        If `reconcile_false_positive` is `True`, the engine may apply additional fixups
        to its state to try to get the rest of the run to just work, assuming the error
        was a false-positive.

        For example, a `tipPhysicallyMissing` error from a `pickUpTip` would normally
        leave the engine state without a tip on the pipette. If `reconcile_false_positive=True`,
        the engine will set the pipette to have that missing tip before continuing, so
        subsequent path planning, aspirates, dispenses, etc. will work as if nothing
        went wrong.
        """
        if reconcile_false_positive:
            state_update = (
                self._state_store.commands.get_state_update_for_false_positive()
            )
        else:
            state_update = StateUpdate()  # Empty/no-op.

        action = self._state_store.commands.validate_action_allowed(
            ResumeFromRecoveryAction(state_update)
        )

        self._action_dispatcher.dispatch(action)

    def add_command(
        self, request: commands.CommandCreate, failed_command_id: Optional[str] = None
    ) -> commands.Command:
        """Add a command to the `ProtocolEngine`'s queue.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The full, newly queued command.

        Raises:
            SetupCommandNotAllowed: the request specified a setup command,
                but the engine was not idle or paused.
            RunStoppedError: the run has been stopped, so no new commands
                may be added.
            CommandNotAllowedError: the request specified a failed command id
                with a non fixit command.
        """
        request = slot_standardization.standardize_command(
            request, self.state_view.config.robot_type
        )

        if failed_command_id and request.intent != commands.CommandIntent.FIXIT:
            raise CommandNotAllowedError(
                "failed command id should be supplied with a FIXIT command."
            )

        command_id = self._model_utils.generate_id()
        if request.intent in (
            commands.CommandIntent.SETUP,
            commands.CommandIntent.FIXIT,
        ):
            request_hash = None
        else:
            request_hash = commands.hash_protocol_command_params(
                create=request,
                last_hash=self._state_store.commands.get_latest_protocol_command_hash(),
            )

        action = self.state_view.commands.validate_action_allowed(
            QueueCommandAction(
                request=request,
                request_hash=request_hash,
                command_id=command_id,
                created_at=self._model_utils.get_timestamp(),
                failed_command_id=failed_command_id,
            )
        )
        self._action_dispatcher.dispatch(action)
        return self._state_store.commands.get(command_id)

    async def wait_for_command(self, command_id: str) -> None:
        """Wait for a command to be completed.

        Will also return if the engine was stopped before it reached the command.
        """
        await self._state_store.wait_for(
            self._state_store.commands.get_command_is_final,
            command_id=command_id,
        )

    async def add_and_execute_command(
        self, request: commands.CommandCreate
    ) -> commands.Command:
        """Add a command to the queue and wait for it to complete.

        The engine must be started by calling `play` before the command will
        execute. You only need to call `play` once.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The command.

            If the command completed, it will be succeeded or failed.

            If the engine was stopped before it reached the command,
            the command will be queued.
        """
        command = self.add_command(request)
        await self.wait_for_command(command.id)
        return self._state_store.commands.get(command.id)

    async def add_and_execute_command_wait_for_recovery(
        self, request: commands.CommandCreate
    ) -> commands.Command:
        """Like `add_and_execute_command()`, except wait for error recovery.

        Unlike `add_and_execute_command()`, if the command fails, this will not
        immediately return the failed command. Instead, if the error is recoverable,
        it will wait until error recovery has completed (e.g. when some other task
        calls `self.resume_from_recovery()`).

        Returns:
            The command.

            If the command completed, it will be succeeded or failed. If it failed
            and then its failure was recovered from, it will still be failed.

            If the engine was stopped before it reached the command,
            the command will be queued.
        """
        queued_command = self.add_command(request)
        await self.wait_for_command(command_id=queued_command.id)
        completed_command = self._state_store.commands.get(queued_command.id)
        await self._state_store.wait_for_not(
            self.state_view.commands.get_recovery_in_progress_for_command,
            queued_command.id,
        )
        return completed_command

    def estop(self) -> None:
        """Signal to the engine that an E-stop event occurred.

        If an estop happens while the robot is moving, lower layers physically stop
        motion and raise the event as an exception, which fails the Protocol Engine
        command. No action from the `ProtocolEngine` caller is needed to handle that.

        However, if an estop happens in between commands, or in the middle of
        a command like `comment` or `waitForDuration` that doesn't access the hardware,
        `ProtocolEngine` needs to be told about it so it can interrupt the command
        and stop executing any more. This method is how to do that.

        This acts roughly like `request_stop()`. After calling this, you should call
        `finish()` with an EStopActivatedError.
        """
        try:
            action = self._state_store.commands.validate_action_allowed(
                StopAction(from_estop=True)
            )
        except Exception:  # todo(mm, 2024-04-16): Catch a more specific type.
            # This is likely called from some hardware API callback that doesn't care
            # about ProtocolEngine lifecycle or what methods are valid to call at what
            # times. So it makes more sense for us to no-op here than to propagate this
            # as an error.
            _log.info(
                "ProtocolEngine cannot handle E-stop event right now. Ignoring it.",
                exc_info=True,
            )
            return
        self._action_dispatcher.dispatch(action)
        # self._queue_worker.cancel() will try to interrupt any ongoing command.
        # Unfortunately, if it's a hardware command, this interruption will race
        # against the E-stop exception propagating up from lower layers. But we need to
        # do this because we want to make sure non-hardware commands, like
        # `waitForDuration`, are also interrupted.
        self._get_queue_worker.cancel()
        # Unlike self.request_stop(), we don't need to do
        # self._hardware_api.cancel_execution_and_running_tasks(). Since this was an
        # E-stop event, the hardware API already knows.

    async def request_stop(self) -> None:
        """Make command execution stop soon.

        This will try to interrupt the ongoing command, if there is one. Future commands
        are canceled. However, by the time this method returns, things may not have
        settled by the time this method returns; the last command may still be
        running.

        After a stop has been requested, the engine cannot be restarted.

        After a stop request, you must still call `finish` to give the engine a chance
        to clean up resources and propagate errors.
        """
        action = self._state_store.commands.validate_action_allowed(StopAction())
        self._action_dispatcher.dispatch(action)
        self._get_queue_worker.cancel()
        if self._hardware_api.is_movement_execution_taskified():
            # We 'taskify' hardware controller movement functions when running protocols
            # that are not backed by the engine. Such runs cannot be stopped by cancelling
            # the queue worker and hence need to be stopped via the execution manager.
            # `cancel_execution_and_running_tasks()` sets the execution manager in a CANCELLED state
            # and cancels the running tasks, which raises an error and gets us out of the
            # run function execution, just like `_queue_worker.cancel()` does for
            # engine-backed runs.
            await self._hardware_api.cancel_execution_and_running_tasks()

    async def wait_until_complete(self) -> None:
        """Wait until there are no more commands to execute.

        If a command encountered a fatal error, it's raised as an exception.
        """
        await self._state_store.wait_for(
            condition=self._state_store.commands.get_all_commands_final
        )
        self._state_store.commands.raise_fatal_command_error()

    async def finish(
        self,
        error: Optional[Exception] = None,
        drop_tips_after_run: bool = True,
        set_run_status: bool = True,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    ) -> None:
        """Finish using the `ProtocolEngine`.

        This does a few things:

        1. It may do post-run actions like homing and dropping tips. This depends on the
           arguments passed as well as heuristics based on the history of the engine.
        2. It waits for the engine to be done controlling the robot's hardware.
        3. It releases internal resources, like background tasks.

        It's safe to call `finish()` multiple times. After you call `finish()`,
        the engine can't be restarted.

        This method should not raise. If any exceptions happened during execution that were not
        properly caught by `ProtocolEngine` internals, or if any exceptions happen during this
        `finish()` call, they should be saved as `.state_view.get_summary().errors`.

        Arguments:
            error: An error that caused the stop, if applicable.
            drop_tips_after_run: Whether to drop tips as part of cleanup.
            set_run_status: Whether to calculate a `success` or `failure` run status.
                If `False`, will set status to `stopped`.
            post_run_hardware_state: The state in which to leave the gantry and motors in
                after the run is over.
        """
        if self._state_store.commands.get_is_stopped_by_estop():
            # This handles the case where the E-stop was pressed while we were *not* in the middle
            # of some hardware interaction that would raise it as an exception. For example, imagine
            # we were paused between two commands, or imagine we were executing a waitForDuration.
            drop_tips_after_run = False
            post_run_hardware_state = PostRunHardwareState.DISENGAGE_IN_PLACE
            if error is None:
                error = EStopActivatedError()

        if error:
            # If the run had an error, check if that error indicates an E-stop.
            # This handles the case where the run was in the middle of some hardware control
            # method and the hardware controller raised an E-stop error from it.
            #
            # To do this, we need to scan all the way through the error tree.
            # By the time E-stop error has gotten to us, it may have been wrapped in other errors,
            # so we need to unwrap them to uncover the E-stop error's inner beauty.
            #
            # We don't use self._hardware_api.get_estop_state() because the E-stop may have been
            # released by the time we get here.
            if isinstance(error, EnumeratedError):
                if code_in_error_tree(
                    root_error=error, code=ErrorCodes.E_STOP_ACTIVATED
                ) or code_in_error_tree(
                    # Request from the hardware team for the v7.0 betas: to help in-house debugging
                    # of pipette overpressure events, leave the pipette where it was like we do
                    # for E-stops.
                    root_error=error,
                    code=ErrorCodes.PIPETTE_OVERPRESSURE,
                ):
                    drop_tips_after_run = False
                    post_run_hardware_state = PostRunHardwareState.DISENGAGE_IN_PLACE

            error_details: Optional[FinishErrorDetails] = FinishErrorDetails(
                error_id=self._model_utils.generate_id(),
                created_at=self._model_utils.get_timestamp(),
                error=error,
            )
        else:
            error_details = None

        self._action_dispatcher.dispatch(
            FinishAction(error_details=error_details, set_run_status=set_run_status)
        )

        # We have a lot of independent things to tear down. If any teardown fails, we want
        # to continue with the rest, to avoid leaking resources or leaving the engine with a broken
        # state. We use an AsyncExitStack to avoid a gigantic try/finally tree. Note that execution
        # order will be backwards because the stack is first-in-last-out.
        exit_stack = AsyncExitStack()
        exit_stack.push_async_callback(self._plugin_starter.stop)  # Last step.
        exit_stack.push_async_callback(
            # Cleanup after hardware halt and reset the hardware controller
            self._hardware_stopper.do_stop_and_recover,
            post_run_hardware_state=post_run_hardware_state,
            drop_tips_after_run=drop_tips_after_run,
        )
        exit_stack.callback(self._door_watcher.stop)

        disengage_before_stopping = (
            False
            if post_run_hardware_state == PostRunHardwareState.STAY_ENGAGED_IN_PLACE
            else True
        )
        # Halt any movements immediately
        exit_stack.push_async_callback(
            self._hardware_stopper.do_halt,
            disengage_before_stopping=disengage_before_stopping,
        )
        exit_stack.push_async_callback(self._get_queue_worker.join)  # First step.
        try:
            # If any teardown steps failed, this will raise something.
            await exit_stack.aclose()
        except Exception as hardware_stopped_exception:
            _log.exception("Exception during post-run finish steps.")
            finish_error_details: Optional[FinishErrorDetails] = FinishErrorDetails(
                error_id=self._model_utils.generate_id(),
                created_at=self._model_utils.get_timestamp(),
                error=hardware_stopped_exception,
            )
        else:
            finish_error_details = None

        self._action_dispatcher.dispatch(
            HardwareStoppedAction(
                completed_at=self._model_utils.get_timestamp(),
                finish_error_details=finish_error_details,
            )
        )

    def add_labware_offset(
        self, request: LabwareOffsetCreate | LegacyLabwareOffsetCreate
    ) -> LabwareOffset:
        """Add a new labware offset and return it.

        The added offset will apply to subsequent `LoadLabwareCommand`s.

        To retrieve offsets later, see `.state_view.labware`.
        """
        internal_request = (
            labware_offset_standardization.standardize_labware_offset_create(
                request,
                self.state_view.config.robot_type,
                self.state_view.addressable_areas.deck_definition,
            )
        )

        labware_offset_id = self._model_utils.generate_id()
        created_at = self._model_utils.get_timestamp()
        self._action_dispatcher.dispatch(
            AddLabwareOffsetAction(
                labware_offset_id=labware_offset_id,
                created_at=created_at,
                request=internal_request,
            )
        )
        return self.state_view.labware.get_labware_offset(
            labware_offset_id=labware_offset_id
        )

    def add_labware_definition(self, definition: LabwareDefinition) -> LabwareUri:
        """Add a labware definition to the state for subsequent labware loads."""
        self._action_dispatcher.dispatch(
            AddLabwareDefinitionAction(definition=definition)
        )
        return self._state_store.labware.get_uri_from_definition(definition)

    def add_liquid(
        self,
        name: str,
        color: Optional[HexColor],
        description: Optional[str],
        id: Optional[str] = None,
    ) -> Liquid:
        """Add a liquid to the state for subsequent liquid loads."""
        if id is None:
            id = self._model_utils.generate_id()

        liquid = Liquid(
            id=id,
            displayName=name,
            description=(description or ""),
            displayColor=color,
        )
        validated_liquid = self._state_store.liquid.validate_liquid_allowed(
            liquid=liquid
        )

        self._action_dispatcher.dispatch(AddLiquidAction(liquid=validated_liquid))
        return validated_liquid

    def add_addressable_area(self, addressable_area_name: str) -> None:
        """Add an addressable area to state."""
        self._action_dispatcher.dispatch(
            AddAddressableAreaAction(addressable_area_name)
        )

    def reset_tips(self, labware_id: str) -> None:
        """Reset the tip state of a given labware."""
        # TODO(mm, 2023-03-10): Safely raise an error if the given labware isn't a
        # tip rack?
        self._action_dispatcher.dispatch(ResetTipsAction(labware_id=labware_id))

    # TODO(mm, 2022-11-10): This is a method on ProtocolEngine instead of a command
    # as a quick hack to support Python protocols. We should consider making this a
    # command, or adding speed parameters to existing commands.
    # https://opentrons.atlassian.net/browse/RCORE-373
    def set_pipette_movement_speed(
        self, pipette_id: str, speed: Optional[float]
    ) -> None:
        """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

        None will use the hardware API's default.
        """
        self._action_dispatcher.dispatch(
            SetPipetteMovementSpeedAction(pipette_id=pipette_id, speed=speed)
        )

    async def use_attached_modules(
        self,
        modules_by_id: Dict[str, HardwareModuleAPI],
    ) -> None:
        """Load attached modules directly into state, without locations."""
        actions = [
            AddModuleAction(
                module_id=module_id,
                serial_number=mod.device_info["serial"],
                definition=self._module_data_provider.get_definition(
                    ModuleModel(mod.model())
                ),
                module_live_data=mod.live_data,
            )
            for module_id, mod in modules_by_id.items()
        ]

        for a in actions:
            self._action_dispatcher.dispatch(a)

    def set_and_start_queue_worker(
        self, command_generator: Callable[[], AsyncGenerator[str, None]]
    ) -> None:
        """Set QueueWorker and start it."""
        assert self._queue_worker is None
        self._queue_worker = create_queue_worker(
            hardware_api=self._hardware_api,
            file_provider=self._file_provider,
            state_store=self._state_store,
            action_dispatcher=self._action_dispatcher,
            command_generator=command_generator,
        )
        self._queue_worker.start()

    def set_error_recovery_policy(self, policy: ErrorRecoveryPolicy) -> None:
        """Replace the run's error recovery policy with a new one."""
        self._action_dispatcher.dispatch(SetErrorRecoveryPolicyAction(policy))


# TODO(tz, 7-12-23): move this to shared data when we dont relay on ErrorOccurrence
def code_in_error_tree(
    root_error: Union[EnumeratedError, ErrorOccurrence], code: ErrorCodes
) -> bool:
    """Check if the specified error code can be found in the given error tree."""
    if isinstance(root_error, ErrorOccurrence):
        # ErrorOccurrence is not the same as the enumerated error exceptions. Check the
        # code by a string value.
        if root_error.errorCode == code.value.code:
            return True
        return any(
            code_in_error_tree(wrapped, code) for wrapped in root_error.wrappedErrors
        )

    # From here we have an exception, can just check the code + recurse to wrapped errors.
    if root_error.code == code:
        return True

    if (
        isinstance(root_error, ProtocolCommandFailedError)
        and root_error.original_error is not None
    ):
        # For this specific EnumeratedError child, we recurse on the original_error field
        # in favor of the general error.wrapping field.
        return code_in_error_tree(root_error.original_error, code)

    if len(root_error.wrapping) == 0:
        return False
    return any(
        code_in_error_tree(wrapped_error, code) for wrapped_error in root_error.wrapping
    )



"""Common pipetting command base models."""

from __future__ import annotations
from typing import Literal, Tuple, TYPE_CHECKING, Optional
import numpy
from typing_extensions import TypedDict
from pydantic import BaseModel, Field

from opentrons_shared_data.errors import ErrorCodes
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence
from opentrons.protocol_engine.types import AspiratedFluid, FluidKind
from opentrons_shared_data.errors.exceptions import PipetteOverpressureError
from .command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.state.update_types import StateUpdate


if TYPE_CHECKING:
    from ..execution.pipetting import PipettingHandler
    from ..resources import ModelUtils
    from ..notes import CommandNoteAdder


DEFAULT_CORRECTION_VOLUME = 0.0
"""Default correction volume (uL) for any aspirate/ dispense volume."""


class PipetteIdMixin(BaseModel):
    """Mixin for command requests that take a pipette ID."""

    pipetteId: str = Field(
        ...,
        description="Identifier of pipette to use for liquid handling.",
    )


class AspirateVolumeMixin(BaseModel):
    """Mixin for the `volume` field of aspirate commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to aspirate, in µL."
        " Must not be greater than the remaining available amount, which depends on"
        " the pipette (see `loadPipette`), its configuration (see `configureForVolume`),"
        " the tip (see `pickUpTip`), and the amount you've aspirated so far."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )
    correctionVolume: Optional[float] = Field(
        None,
        description="The correction volume in uL.",
    )


class DispenseVolumeMixin(BaseModel):
    """Mixin for the `volume` field of dispense commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to dispense, in µL."
        " Must not be greater than the currently aspirated volume."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )
    correctionVolume: Optional[float] = Field(
        None,
        description="The correction volume in uL.",
    )


class FlowRateMixin(BaseModel):
    """Mixin for command requests that take a flow rate."""

    flowRate: float = Field(
        ..., description="Speed in µL/s configured for the pipette", gt=0
    )


class BaseLiquidHandlingResult(BaseModel):
    """Base properties of a liquid handling result."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL handled in the operation.",
        ge=0,
    )


class EmptyResult(BaseModel):
    """A result with no data."""


class ErrorLocationInfo(TypedDict):
    """Holds a retry location for in-place error recovery.

    This is appropriate to pass to a `moveToCoordinates` command,
    assuming the pipette has not been configured with a different nozzle layout
    in the meantime.
    """

    retryLocation: Tuple[float, float, float]


class OverpressureError(ErrorOccurrence):
    """Returned when sensors detect an overpressure error while moving liquid.

    The pipette plunger motion is stopped at the point of the error.

    The next thing to move the plunger must account for the robot not having a valid
    estimate of its position. It should be a `home`, `unsafe/updatePositionEstimators`,
    `unsafe/dropTipInPlace`, or `unsafe/blowOutInPlace`.
    """

    isDefined: bool = True

    errorType: Literal["overpressure"] = "overpressure"

    errorCode: str = ErrorCodes.PIPETTE_OVERPRESSURE.value.code
    detail: str = ErrorCodes.PIPETTE_OVERPRESSURE.value.detail

    errorInfo: ErrorLocationInfo


class LiquidNotFoundError(ErrorOccurrence):
    """Returned when no liquid is detected during the liquid probe process/move.

    After a failed probing, the pipette returns to the process start position.
    """

    isDefined: bool = True

    errorType: Literal["liquidNotFound"] = "liquidNotFound"

    errorCode: str = ErrorCodes.PIPETTE_LIQUID_NOT_FOUND.value.code
    detail: str = ErrorCodes.PIPETTE_LIQUID_NOT_FOUND.value.detail


class TipPhysicallyAttachedError(ErrorOccurrence):
    """Returned when sensors determine that a tip remains on the pipette after a drop attempt.

    The pipette will act as if the tip was not dropped. So, you won't be able to pick
    up a new tip without dropping the current one, and movement commands will assume
    there is a tip hanging off the bottom of the pipette.
    """

    isDefined: bool = True

    errorType: Literal["tipPhysicallyAttached"] = "tipPhysicallyAttached"

    errorCode: str = ErrorCodes.TIP_DROP_FAILED.value.code
    detail: str = ErrorCodes.TIP_DROP_FAILED.value.detail

    errorInfo: ErrorLocationInfo


async def prepare_for_aspirate(
    pipette_id: str,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
    location_if_error: ErrorLocationInfo,
) -> SuccessData[EmptyResult] | DefinedErrorData[OverpressureError]:
    """Execute pipetting.prepare_for_aspirate, handle errors, and marshal success."""
    try:
        await pipetting.prepare_for_aspirate(pipette_id)
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate().set_fluid_unknown(pipette_id=pipette_id),
        )
    else:
        return SuccessData(
            public=EmptyResult(),
            state_update=StateUpdate()
            .set_fluid_empty(pipette_id=pipette_id)
            .set_pipette_ready_to_aspirate(
                pipette_id=pipette_id, ready_to_aspirate=True
            ),
        )


async def aspirate_in_place(
    pipette_id: str,
    volume: float,
    flow_rate: float,
    correction_volume: float,
    location_if_error: ErrorLocationInfo,
    command_note_adder: CommandNoteAdder,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> SuccessData[BaseLiquidHandlingResult] | DefinedErrorData[OverpressureError]:
    """Execute an aspirate in place micro-operation."""
    try:
        volume_aspirated = await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=volume,
            flow_rate=flow_rate,
            command_note_adder=command_note_adder,
            correction_volume=correction_volume,
        )
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate().set_fluid_unknown(pipette_id=pipette_id),
        )
    else:
        return SuccessData(
            public=BaseLiquidHandlingResult(
                volume=volume_aspirated,
            ),
            state_update=StateUpdate().set_fluid_aspirated(
                pipette_id=pipette_id,
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=volume_aspirated),
            ),
        )


async def aspirate_while_tracking(
    pipette_id: str,
    labware_id: str,
    well_name: str,
    volume: float,
    flow_rate: float,
    location_if_error: ErrorLocationInfo,
    command_note_adder: CommandNoteAdder,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> SuccessData[BaseLiquidHandlingResult] | DefinedErrorData[OverpressureError]:
    """Execute an aspirate while tracking microoperation."""
    try:
        volume_aspirated = await pipetting.aspirate_while_tracking(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            volume=volume,
            flow_rate=flow_rate,
            command_note_adder=command_note_adder,
        )
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate().set_fluid_unknown(pipette_id=pipette_id),
        )
    else:
        return SuccessData(
            public=BaseLiquidHandlingResult(
                volume=volume_aspirated,
            ),
            state_update=StateUpdate().set_fluid_aspirated(
                pipette_id=pipette_id,
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=volume_aspirated),
            ),
        )


async def dispense_while_tracking(
    pipette_id: str,
    labware_id: str,
    well_name: str,
    volume: float,
    flow_rate: float,
    push_out: float | None,
    location_if_error: ErrorLocationInfo,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> SuccessData[BaseLiquidHandlingResult] | DefinedErrorData[OverpressureError]:
    """Execute an dispense while tracking microoperation."""
    # The current volume won't be none since it passed validation
    current_volume = (
        pipetting.get_state_view().pipettes.get_aspirated_volume(pipette_id) or 0.0
    )
    is_full_dispense = bool(numpy.isclose(current_volume - volume, 0))
    ready = push_out == 0 if push_out is not None else not is_full_dispense
    try:
        volume_dispensed = await pipetting.dispense_while_tracking(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            volume=volume,
            flow_rate=flow_rate,
            push_out=push_out,
            is_full_dispense=is_full_dispense,
        )
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate()
            .set_fluid_unknown(pipette_id=pipette_id)
            .set_pipette_ready_to_aspirate(
                pipette_id=pipette_id, ready_to_aspirate=False
            ),
        )
    else:
        return SuccessData(
            public=BaseLiquidHandlingResult(
                volume=volume_dispensed,
            ),
            state_update=StateUpdate()
            .set_fluid_ejected(
                pipette_id=pipette_id,
                volume=volume_dispensed,
            )
            .set_pipette_ready_to_aspirate(
                pipette_id=pipette_id, ready_to_aspirate=ready
            ),
        )


async def dispense_in_place(
    pipette_id: str,
    volume: float,
    flow_rate: float,
    push_out: float | None,
    correction_volume: float,
    location_if_error: ErrorLocationInfo,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> SuccessData[BaseLiquidHandlingResult] | DefinedErrorData[OverpressureError]:
    """Dispense-in-place as a micro-operation."""
    # The current volume won't be none since it passed validation
    current_volume = (
        pipetting.get_state_view().pipettes.get_aspirated_volume(pipette_id) or 0.0
    )
    is_full_dispense = bool(numpy.isclose(current_volume - volume, 0))
    ready: bool = push_out == 0 if push_out is not None else not is_full_dispense
    try:
        volume = await pipetting.dispense_in_place(
            pipette_id=pipette_id,
            volume=volume,
            flow_rate=flow_rate,
            push_out=push_out,
            is_full_dispense=is_full_dispense,
            correction_volume=correction_volume,
        )
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate()
            .set_fluid_unknown(pipette_id=pipette_id)
            .set_pipette_ready_to_aspirate(
                pipette_id=pipette_id, ready_to_aspirate=False
            ),
        )
    else:
        return SuccessData(
            public=BaseLiquidHandlingResult(volume=volume),
            state_update=StateUpdate()
            .set_fluid_ejected(pipette_id=pipette_id, volume=volume)
            .set_pipette_ready_to_aspirate(
                pipette_id=pipette_id, ready_to_aspirate=ready
            ),
        )


async def blow_out_in_place(
    pipette_id: str,
    flow_rate: float,
    location_if_error: ErrorLocationInfo,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> SuccessData[EmptyResult] | DefinedErrorData[OverpressureError]:
    """Execute a blow-out-in-place micro-operation."""
    try:
        await pipetting.blow_out_in_place(pipette_id=pipette_id, flow_rate=flow_rate)
    except PipetteOverpressureError as e:
        return DefinedErrorData(
            public=OverpressureError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
                errorInfo=location_if_error,
            ),
            state_update=StateUpdate().set_fluid_unknown(pipette_id=pipette_id),
        )
    else:
        return SuccessData(
            public=EmptyResult(),
            state_update=StateUpdate().set_fluid_empty(pipette_id=pipette_id),
        )


async def increase_evo_disp_count(pipette_id: str, pipetting: PipettingHandler) -> None:
    """Tell a pipette to increase it's evo-tip-dispense-count in eeprom."""
    await pipetting.increase_evo_disp_count(pipette_id)



"""Protocol Engine CommandStore sub-state."""
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, List, Optional

from opentrons.ordered_set import OrderedSet
from opentrons.protocol_engine.errors.exceptions import CommandDoesNotExistError

from ..commands import Command, CommandStatus, CommandIntent


@dataclass(frozen=True)
class CommandEntry:
    """A command entry in state, including its index in the list."""

    command: Command
    index: int


@dataclass  # dataclass for __eq__() autogeneration.
class CommandHistory:
    """Provides O(1) amortized access to commands of interest."""

    _all_command_ids: List[str]
    """All command IDs, in insertion order."""

    _all_failed_command_ids: List[str]
    """All failed command IDs, in insertion order."""

    _all_command_ids_but_fixit_command_ids: List[str]
    """All command IDs besides fixit command intents, in insertion order."""

    _commands_by_id: Dict[str, CommandEntry]
    """All command resources, in insertion order, mapped by their unique IDs."""

    _queued_command_ids: OrderedSet[str]
    """The IDs of queued commands, in FIFO order"""

    _queued_setup_command_ids: OrderedSet[str]
    """The IDs of queued setup commands, in FIFO order"""

    _queued_fixit_command_ids: OrderedSet[str]
    """The IDs of queued fixit commands, in FIFO order"""

    _running_command_id: Optional[str]
    """The ID of the currently running command, if any"""

    _most_recently_completed_command_id: Optional[str]
    """ID of the most recent command that SUCCEEDED or FAILED, if any"""

    def __init__(self) -> None:
        self._all_command_ids = []
        self._all_failed_command_ids = []
        self._all_command_ids_but_fixit_command_ids = []
        self._queued_command_ids = OrderedSet()
        self._queued_setup_command_ids = OrderedSet()
        self._queued_fixit_command_ids = OrderedSet()
        self._commands_by_id = OrderedDict()
        self._running_command_id = None
        self._most_recently_completed_command_id = None

    def length(self) -> int:
        """Get the length of all elements added to the history."""
        return len(self._commands_by_id)

    def has(self, command_id: str) -> bool:
        """Returns whether a command is in the history."""
        return command_id in self._commands_by_id

    def get(self, command_id: str) -> CommandEntry:
        """Get a command entry if present, otherwise raise an exception."""
        try:
            return self._commands_by_id[command_id]
        except KeyError:
            raise CommandDoesNotExistError(f"Command {command_id} does not exist")

    def get_next(self, command_id: str) -> Optional[CommandEntry]:
        """Get the command which follows the command associated with the given ID, if any."""
        index = self.get(command_id).index
        try:
            return self._commands_by_id[self._all_command_ids[index + 1]]
        except KeyError:
            raise CommandDoesNotExistError(f"Command {command_id} does not exist")
        except IndexError:
            return None

    def get_prev(self, command_id: str) -> Optional[CommandEntry]:
        """Get the command which precedes the command associated with the given ID, if any.

        Returns None if the command_id corresponds to the first element in the history.
        """
        index = self.get(command_id).index
        try:
            prev_command = self._commands_by_id[self._all_command_ids[index - 1]]
            return prev_command if index != 0 else None
        except KeyError:
            raise CommandDoesNotExistError(f"Command {command_id} does not exist")
        except IndexError:
            return None

    def get_all_commands(self) -> List[Command]:
        """Get all commands."""
        return [
            self._commands_by_id[command_id].command
            for command_id in self._all_command_ids
        ]

    def get_all_failed_commands(self) -> List[Command]:
        """Get all failed commands."""
        return [
            self._commands_by_id[command_id].command
            for command_id in self._all_failed_command_ids
        ]

    def get_filtered_command_ids(self, include_fixit_commands: bool) -> List[str]:
        """Get all fixit command IDs."""
        if include_fixit_commands:
            return self._all_command_ids
        else:
            return self._all_command_ids_but_fixit_command_ids

    def get_all_ids(self) -> List[str]:
        """Get all command IDs."""
        return self._all_command_ids

    def get_slice(
        self, start: int, stop: int, command_ids: Optional[list[str]] = None
    ) -> List[Command]:
        """Get a list of commands between start and stop.""" """Get a list of commands between start and stop."""
        commands = self._all_command_ids[start:stop]
        selected_command_ids = (
            command_ids if command_ids is not None else self._all_command_ids
        )
        commands = selected_command_ids[start:stop]
        return [self._commands_by_id[command].command for command in commands]

    def get_tail_command(self) -> Optional[CommandEntry]:
        """Get the command most recently added."""
        if self._commands_by_id:
            return next(reversed(self._commands_by_id.values()))
        else:
            return None

    def get_most_recently_completed_command(self) -> Optional[CommandEntry]:
        """Get the command most recently marked as SUCCEEDED or FAILED."""
        if self._most_recently_completed_command_id is not None:
            return self._commands_by_id[self._most_recently_completed_command_id]
        else:
            return None

    def get_running_command(self) -> Optional[CommandEntry]:
        """Get the command currently running, if any."""
        if self._running_command_id is None:
            return None
        else:
            return self._commands_by_id[self._running_command_id]

    def get_queue_ids(self) -> OrderedSet[str]:
        """Get the IDs of all queued protocol commands, in FIFO order."""
        return self._queued_command_ids

    def get_setup_queue_ids(self) -> OrderedSet[str]:
        """Get the IDs of all queued setup commands, in FIFO order."""
        return self._queued_setup_command_ids

    def get_fixit_queue_ids(self) -> OrderedSet[str]:
        """Get the IDs of all queued fixit commands, in FIFO order."""
        return self._queued_fixit_command_ids

    def append_queued_command(self, command: Command) -> None:
        """Validate and mark a command as queued in the command history."""
        assert command.status == CommandStatus.QUEUED
        assert not self.has(command.id)

        next_index = self.length()
        updated_command = CommandEntry(
            index=next_index,
            command=command,
        )
        self._add(command.id, updated_command)

        if command.intent == CommandIntent.SETUP:
            self._add_to_setup_queue(command.id)
        elif command.intent == CommandIntent.FIXIT:
            self._add_to_fixit_queue(command.id)
        else:
            self._add_to_queue(command.id)

    def set_command_running(self, command: Command) -> None:
        """Validate and mark a command as running in the command history."""
        prev_entry = self.get(command.id)

        assert prev_entry.command.status == CommandStatus.QUEUED
        assert command.status == CommandStatus.RUNNING

        self._add(
            command.id,
            CommandEntry(index=prev_entry.index, command=command),
        )

        assert self.get_running_command() is None
        self._set_running_command_id(command.id)

        self._remove_queue_id(command.id)
        self._remove_setup_queue_id(command.id)
        self._remove_fixit_queue_id(command.id)

    def set_command_succeeded(self, command: Command) -> None:
        """Validate and mark a command as succeeded in the command history."""
        prev_entry = self.get(command.id)
        assert prev_entry.command.status == CommandStatus.RUNNING
        assert command.status == CommandStatus.SUCCEEDED

        self._add(
            command.id,
            CommandEntry(
                index=prev_entry.index,
                command=command,
            ),
        )

        running_command_entry = self.get_running_command()
        assert running_command_entry is not None
        assert running_command_entry.command.id == command.id
        self._set_running_command_id(None)

        self._remove_queue_id(command.id)
        self._remove_setup_queue_id(command.id)
        self._set_most_recently_completed_command_id(command.id)

    def set_command_failed(self, command: Command) -> None:
        """Validate and mark a command as failed in the command history."""
        prev_entry = self.get(command.id)
        assert (
            prev_entry.command.status == CommandStatus.RUNNING
            or prev_entry.command.status == CommandStatus.QUEUED
        )
        assert command.status == CommandStatus.FAILED

        index = self.get(command.id).index
        self._add(
            command_id=command.id,
            command_entry=CommandEntry(index=index, command=command),
        )

        running_command_entry = self.get_running_command()
        if (
            running_command_entry is not None
            and running_command_entry.command.id == command.id
        ):
            self._set_running_command_id(None)

        self._remove_queue_id(command.id)
        self._remove_setup_queue_id(command.id)
        self._set_most_recently_completed_command_id(command.id)
        self._all_failed_command_ids.append(command.id)

    def _add(self, command_id: str, command_entry: CommandEntry) -> None:
        """Create or update a command entry."""
        if command_id not in self._commands_by_id:
            self._all_command_ids.append(command_id)
            if command_entry.command.intent != CommandIntent.FIXIT:
                self._all_command_ids_but_fixit_command_ids.append(command_id)
        self._commands_by_id[command_id] = command_entry

    def _add_to_queue(self, command_id: str) -> None:
        """Add new ID to the queued."""
        self._queued_command_ids.add(command_id)

    def _add_to_setup_queue(self, command_id: str) -> None:
        """Add a new ID to the queued setup."""
        self._queued_setup_command_ids.add(command_id)

    def _add_to_fixit_queue(self, command_id: str) -> None:
        """Add a new ID to the queued fixit."""
        self._queued_fixit_command_ids.add(command_id)

    def _remove_queue_id(self, command_id: str) -> None:
        """Remove a specific command from the queued command ids structure."""
        self._queued_command_ids.discard(command_id)

    def _remove_setup_queue_id(self, command_id: str) -> None:
        """Remove a specific command from the queued setup command ids structure."""
        self._queued_setup_command_ids.discard(command_id)

    def _remove_fixit_queue_id(self, command_id: str) -> None:
        """Remove a specific command from the queued fixit command ids structure."""
        self._queued_fixit_command_ids.discard(command_id)

    def _set_most_recently_completed_command_id(self, command_id: str) -> None:
        """Set the ID of the most recently dequeued command."""
        self._most_recently_completed_command_id = command_id

    def _set_running_command_id(self, command_id: Optional[str]) -> None:
        """Set the ID of the currently running command."""
        self._running_command_id = command_id
